// sciplot - a modern C++ scientific plotting library powered by gnuplot
// https://github.com/sciplot/sciplot
//
// Licensed under the MIT License <http://opensource.org/licenses/MIT>.
//
// Copyright (c) 2018-2021 Allan Leal
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#pragma once

// C++ includes
#include <map>
#include <string>

namespace sciplot {

/// Gnuplot color palettes for sciplot adapted from https://github.com/Gnuplotting/gnuplot-palettes
const std::map<std::string, std::string> palettes = {
    { "accent", "# line styles for ColorBrewer Accent\n# for use with qualitative/categorical data\n# provides 8 colors, 4 pale and 4 saturated\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#7FC97F' # pale green\nset style line 2 lt 1 lc rgb '#BEAED4' # pale purple\nset style line 3 lt 1 lc rgb '#FDC086' # pale orange\nset style line 4 lt 1 lc rgb '#FFFF99' # pale yellow\nset style line 5 lt 1 lc rgb '#386CB0' # blue\nset style line 6 lt 1 lc rgb '#F0027F' # magenta\nset style line 7 lt 1 lc rgb '#BF5B17' # brown\nset style line 8 lt 1 lc rgb '#666666' # grey\n\n# palette\nset palette maxcolors 8\nset palette defined ( 0 '#7FC97F',\\\n    \t    \t      1 '#BEAED4',\\\n\t\t      2 '#FDC086',\\\n\t\t      3 '#FFFF99',\\\n\t\t      4 '#386CB0',\\\n\t\t      5 '#F0027F',\\\n\t\t      6 '#BF5B17',\\\n\t\t      7 '#666666' )\n" },
    { "blues", "# line styles for ColorBrewer Blues\n# for use with sequential data\n# provides 8 blue colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#F7FBFF' # very light blue\nset style line 2 lt 1 lc rgb '#DEEBF7' # \nset style line 3 lt 1 lc rgb '#C6DBEF' # \nset style line 4 lt 1 lc rgb '#9ECAE1' # light blue\nset style line 5 lt 1 lc rgb '#6BAED6' # \nset style line 6 lt 1 lc rgb '#4292C6' # medium blue\nset style line 7 lt 1 lc rgb '#2171B5' #\nset style line 8 lt 1 lc rgb '#084594' # dark blue\n\n# palette\nset palette defined ( 0 '#F7FBFF',\\\n    \t    \t      1 '#DEEBF7',\\\n\t\t      2 '#C6DBEF',\\\n\t\t      3 '#9ECAE1',\\\n\t\t      4 '#6BAED6',\\\n\t\t      5 '#4292C6',\\\n\t\t      6 '#2171B5',\\\n\t\t      7 '#084594' )\n" },
    { "brbg", "# line styles for ColorBrewer BrBG\n# for use with divering data\n# provides 8 colors with brown low, white middle, and blue-green high\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#8C510A' # dark brown\nset style line 2 lt 1 lc rgb '#BF812D' # medium brown\nset style line 3 lt 1 lc rgb '#DFC27D' # \nset style line 4 lt 1 lc rgb '#F6E8C3' # pale brown\nset style line 5 lt 1 lc rgb '#C7EAE5' # pale blue-green\nset style line 6 lt 1 lc rgb '#80CDC1' # \nset style line 7 lt 1 lc rgb '#35978F' # medium blue-green\nset style line 8 lt 1 lc rgb '#01665E' # dark blue-green\n\n# palette\nset palette defined ( 0 '#8C510A',\\\n    \t    \t      1 '#BF812D',\\\n\t\t      2 '#DFC27D',\\\n\t\t      3 '#F6E8C3',\\\n\t\t      4 '#C7EAE5',\\\n\t\t      5 '#80CDC1',\\\n\t\t      6 '#35978F',\\\n\t\t      7 '#01665E' )\n" },
    { "bugn", "# line styles for ColorBrewer BuGn\n# for use with sequential data\n# provides 8 blue-green colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#F7FCFD' # very light blue-green\nset style line 2 lt 1 lc rgb '#E5F5F9' # \nset style line 3 lt 1 lc rgb '#CCECE6' # \nset style line 4 lt 1 lc rgb '#99D8C9' # light blue-green\nset style line 5 lt 1 lc rgb '#66C2A4' # \nset style line 6 lt 1 lc rgb '#41AE76' # medium blue-green\nset style line 7 lt 1 lc rgb '#238B45' #\nset style line 8 lt 1 lc rgb '#005824' # dark blue-green\n\n# palette\nset palette defined ( 0 '#F7FCFD',\\\n    \t    \t      1 '#E5F5F9',\\\n\t\t      2 '#CCECE6',\\\n\t\t      3 '#99D8C9',\\\n\t\t      4 '#66C2A4',\\\n\t\t      5 '#41AE76',\\\n\t\t      6 '#238B45',\\\n\t\t      7 '#005824' )\n" },
    { "bupu", "# line styles for ColorBrewer BuPu\n# for use with sequential data\n# provides 8 blue-purple colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#F7FCFD' # very light blue-purple\nset style line 2 lt 1 lc rgb '#E0ECF4' # \nset style line 3 lt 1 lc rgb '#BFD3E6' # \nset style line 4 lt 1 lc rgb '#9EBCDA' # light blue-purple\nset style line 5 lt 1 lc rgb '#8C96C6' # \nset style line 6 lt 1 lc rgb '#8C6BB1' # medium blue-purple\nset style line 7 lt 1 lc rgb '#88419D' #\nset style line 8 lt 1 lc rgb '#6E016B' # dark blue-purple\n\n# palette\nset palette defined ( 0 '#F7FCFD',\\\n    \t    \t      1 '#E0ECF4',\\\n\t\t      2 '#BFD3E6',\\\n\t\t      3 '#9EBCDA',\\\n\t\t      4 '#8C96C6',\\\n\t\t      5 '#8C6BB1',\\\n\t\t      6 '#88419D',\\\n\t\t      7 '#6E016B' )\n" },
    { "chromajs", "# line styles for chromajs\n# see:\n# http://gka.github.io/palettes/#colors=lightyellow,orangered,deeppink,darkred|steps=7|bez=1|coL=1\n# https://vis4.net/blog/posts/mastering-multi-hued-color-scales/\n\n# line styles\nset style line 1 lt 1 lc rgb '#ffffe0' # light yellow\nset style line 2 lt 1 lc rgb '#ffdfb8' #\nset style line 3 lt 1 lc rgb '#ffbc94' # light orange\nset style line 4 lt 1 lc rgb '#ff9777' #\nset style line 5 lt 1 lc rgb '#ff6962' # light red\nset style line 6 lt 1 lc rgb '#ee4256' #\nset style line 7 lt 1 lc rgb '#d21f47' # red\nset style line 8 lt 1 lc rgb '#b0062c' #\nset style line 9 lt 1 lc rgb '#8b0000' # dark red\n\n# palette\nset palette maxcolors 9\nset palette defined ( 0 '#ffffe0',\\\n    \t    \t      1 '#ffdfb8',\\\n        \t\t      2 '#ffbc94',\\\n\t\t              3 '#ff9777',\\\n\t\t              4 '#ff6962',\\\n\t\t              5 '#ee4256',\\\n\t\t              6 '#d21f47',\\\n\t\t              7 '#b0062c',\\\n                      8 '#8b0000')\n" },
    { "dark2", "# line styles for ColorBrewer Dark2\n# for use with qualitative/categorical data\n# provides 8 dark colors based on Set2\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#1B9E77' # dark teal\nset style line 2 lt 1 lc rgb '#D95F02' # dark orange\nset style line 3 lt 1 lc rgb '#7570B3' # dark lilac\nset style line 4 lt 1 lc rgb '#E7298A' # dark magenta\nset style line 5 lt 1 lc rgb '#66A61E' # dark lime green\nset style line 6 lt 1 lc rgb '#E6AB02' # dark banana\nset style line 7 lt 1 lc rgb '#A6761D' # dark tan\nset style line 8 lt 1 lc rgb '#666666' # dark gray\n\n# palette\nset palette maxcolors 8\nset palette defined ( 0 '#1B9E77',\\\n    \t    \t      1 '#D95F02',\\\n\t\t      2 '#7570B3',\\\n\t\t      3 '#E7298A',\\\n\t\t      4 '#66A61E',\\\n\t\t      5 '#E6AB02',\\\n\t\t      6 '#A6761D',\\\n\t\t      7 '#666666' )\n" },
    { "gnbu", "# line styles for ColorBrewer GnBu\n# for use with sequential data\n# provides 8 green-blue colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#F7FCF0' # very light green-blue\nset style line 2 lt 1 lc rgb '#E0F3DB' # \nset style line 3 lt 1 lc rgb '#CCEBC5' # \nset style line 4 lt 1 lc rgb '#A8DDB5' # light green-blue\nset style line 5 lt 1 lc rgb '#7BCCC4' # \nset style line 6 lt 1 lc rgb '#4EB3D3' # medium green-blue\nset style line 7 lt 1 lc rgb '#2B8CBE' #\nset style line 8 lt 1 lc rgb '#08589E' # dark green-blue\n\n# palette\nset palette defined ( 0 '#F7FCF0',\\\n    \t    \t      1 '#E0F3DB',\\\n\t\t      2 '#CCEBC5',\\\n\t\t      3 '#A8DDB5',\\\n\t\t      4 '#7BCCC4',\\\n\t\t      5 '#4EB3D3',\\\n\t\t      6 '#2B8CBE',\\\n\t\t      7 '#08589E' )\n" },
    { "gnpu", "# line styles\nset style line  1 lt 1 lc rgb '#396353' # green\nset style line  2 lt 1 lc rgb '#0db14b' # \nset style line  3 lt 1 lc rgb '#6dc067' # \nset style line  4 lt 1 lc rgb '#abd69b' #\nset style line  5 lt 1 lc rgb '#daeac1' # \nset style line  6 lt 1 lc rgb '#dfcce4' #\nset style line  7 lt 1 lc rgb '#c7b2d6' #\nset style line  8 lt 1 lc rgb '#9474b4' #\nset style line  9 lt 1 lc rgb '#754098' #\nset style line 10 lt 1 lc rgb '#504971' # purple\n\n# palette\nset palette defined ( \\\n    1 '#396353', \\\n    2 '#0db14b', \\\n    3 '#6dc067', \\\n    4 '#abd69b', \\\n    5 '#daeac1', \\\n    6 '#dfcce4', \\\n    7 '#c7b2d6', \\\n    8 '#9474b4', \\\n    9 '#754098', \\\n   10 '#504971')\n" },
    { "greens", "# line styles for ColorBrewer Greens\n# for use with sequential data\n# provides 8 green colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#F7FCF5' # very light green\nset style line 2 lt 1 lc rgb '#E5F5E0' # \nset style line 3 lt 1 lc rgb '#C7E9C0' # \nset style line 4 lt 1 lc rgb '#A1D99B' # light green\nset style line 5 lt 1 lc rgb '#74C476' # \nset style line 6 lt 1 lc rgb '#41AB5D' # medium green\nset style line 7 lt 1 lc rgb '#238B45' #\nset style line 8 lt 1 lc rgb '#005A32' # dark green\n\n# palette\nset palette defined ( 0 '#F7FCF5',\\\n    \t    \t      1 '#E5F5E0',\\\n\t\t      2 '#C7E9C0',\\\n\t\t      3 '#A1D99B',\\\n\t\t      4 '#74C476',\\\n\t\t      5 '#41AB5D',\\\n\t\t      6 '#238B45',\\\n\t\t      7 '#005A32' )\n" },
    { "greys", "# line styles for ColorBrewer Greys\n# for use with sequential data\n# provides 8 grey colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#FFFFFF' # white\nset style line 2 lt 1 lc rgb '#F0F0F0' # \nset style line 3 lt 1 lc rgb '#D9D9D9' # \nset style line 4 lt 1 lc rgb '#BDBDBD' # light grey\nset style line 5 lt 1 lc rgb '#969696' # \nset style line 6 lt 1 lc rgb '#737373' # medium grey\nset style line 7 lt 1 lc rgb '#525252' #\nset style line 8 lt 1 lc rgb '#252525' # dark grey\n\n# palette\nset palette defined ( 0 '#FFFFFF',\\\n    \t    \t      1 '#F0F0F0',\\\n\t\t      2 '#D9D9D9',\\\n\t\t      3 '#BDBDBD',\\\n\t\t      4 '#969696',\\\n\t\t      5 '#737373',\\\n\t\t      6 '#525252',\\\n\t\t      7 '#252525' )\n" },
    { "inferno", "# New matplotlib colormaps by Nathaniel J. Smith, Stefan van der Walt,\n# and (in the case of viridis) Eric Firing.\n#\n# This file and the colormaps in it are released under the CC0 license /\n# public domain dedication. We would appreciate credit if you use or\n# redistribute these colormaps, but do not impose any legal restrictions.\n#\n# To the extent possible under law, the persons who associated CC0 with\n# mpl-colormaps have waived all copyright and related or neighboring rights\n# to mpl-colormaps.\n#\n# You should have received a copy of the CC0 legalcode along with this\n# work.  If not, see <http://creativecommons.org/publicdomain/zero/1.0/>.\n\n#https://github.com/BIDS/colormap/blob/master/colormaps.py\n\n\n# line styles\nset style line  1 lt 1 lc rgb '#000004' # black\nset style line  2 lt 1 lc rgb '#1f0c48' # dark purple\nset style line  3 lt 1 lc rgb '#550f6d' # dark purple\nset style line  4 lt 1 lc rgb '#88226a' # purple\nset style line  5 lt 1 lc rgb '#a83655' # red-magenta\nset style line  6 lt 1 lc rgb '#e35933' # red\nset style line  7 lt 1 lc rgb '#f9950a' # orange\nset style line  8 lt 1 lc rgb '#f8c932' # yellow-orange\nset style line  9 lt 1 lc rgb '#fcffa4' # light yellow\n\n\n# palette\nset palette defined (\\\n0   0.001462 0.000466 0.013866,\\\n1   0.002267 0.001270 0.018570,\\\n2   0.003299 0.002249 0.024239,\\\n3   0.004547 0.003392 0.030909,\\\n4   0.006006 0.004692 0.038558,\\\n5   0.007676 0.006136 0.046836,\\\n6   0.009561 0.007713 0.055143,\\\n7   0.011663 0.009417 0.063460,\\\n8   0.013995 0.011225 0.071862,\\\n9   0.016561 0.013136 0.080282,\\\n10  0.019373 0.015133 0.088767,\\\n11  0.022447 0.017199 0.097327,\\\n12  0.025793 0.019331 0.105930,\\\n13  0.029432 0.021503 0.114621,\\\n14  0.033385 0.023702 0.123397,\\\n15  0.037668 0.025921 0.132232,\\\n16  0.042253 0.028139 0.141141,\\\n17  0.046915 0.030324 0.150164,\\\n18  0.051644 0.032474 0.159254,\\\n19  0.056449 0.034569 0.168414,\\\n20  0.061340 0.036590 0.177642,\\\n21  0.066331 0.038504 0.186962,\\\n22  0.071429 0.040294 0.196354,\\\n23  0.076637 0.041905 0.205799,\\\n24  0.081962 0.043328 0.215289,\\\n25  0.087411 0.044556 0.224813,\\\n26  0.092990 0.045583 0.234358,\\\n27  0.098702 0.046402 0.243904,\\\n28  0.104551 0.047008 0.253430,\\\n29  0.110536 0.047399 0.262912,\\\n30  0.116656 0.047574 0.272321,\\\n31  0.122908 0.047536 0.281624,\\\n32  0.129285 0.047293 0.290788,\\\n33  0.135778 0.046856 0.299776,\\\n34  0.142378 0.046242 0.308553,\\\n35  0.149073 0.045468 0.317085,\\\n36  0.155850 0.044559 0.325338,\\\n37  0.162689 0.043554 0.333277,\\\n38  0.169575 0.042489 0.340874,\\\n39  0.176493 0.041402 0.348111,\\\n40  0.183429 0.040329 0.354971,\\\n41  0.190367 0.039309 0.361447,\\\n42  0.197297 0.038400 0.367535,\\\n43  0.204209 0.037632 0.373238,\\\n44  0.211095 0.037030 0.378563,\\\n45  0.217949 0.036615 0.383522,\\\n46  0.224763 0.036405 0.388129,\\\n47  0.231538 0.036405 0.392400,\\\n48  0.238273 0.036621 0.396353,\\\n49  0.244967 0.037055 0.400007,\\\n50  0.251620 0.037705 0.403378,\\\n51  0.258234 0.038571 0.406485,\\\n52  0.264810 0.039647 0.409345,\\\n53  0.271347 0.040922 0.411976,\\\n54  0.277850 0.042353 0.414392,\\\n55  0.284321 0.043933 0.416608,\\\n56  0.290763 0.045644 0.418637,\\\n57  0.297178 0.047470 0.420491,\\\n58  0.303568 0.049396 0.422182,\\\n59  0.309935 0.051407 0.423721,\\\n60  0.316282 0.053490 0.425116,\\\n61  0.322610 0.055634 0.426377,\\\n62  0.328921 0.057827 0.427511,\\\n63  0.335217 0.060060 0.428524,\\\n64  0.341500 0.062325 0.429425,\\\n65  0.347771 0.064616 0.430217,\\\n66  0.354032 0.066925 0.430906,\\\n67  0.360284 0.069247 0.431497,\\\n68  0.366529 0.071579 0.431994,\\\n69  0.372768 0.073915 0.432400,\\\n70  0.379001 0.076253 0.432719,\\\n71  0.385228 0.078591 0.432955,\\\n72  0.391453 0.080927 0.433109,\\\n73  0.397674 0.083257 0.433183,\\\n74  0.403894 0.085580 0.433179,\\\n75  0.410113 0.087896 0.433098,\\\n76  0.416331 0.090203 0.432943,\\\n77  0.422549 0.092501 0.432714,\\\n78  0.428768 0.094790 0.432412,\\\n79  0.434987 0.097069 0.432039,\\\n80  0.441207 0.099338 0.431594,\\\n81  0.447428 0.101597 0.431080,\\\n82  0.453651 0.103848 0.430498,\\\n83  0.459875 0.106089 0.429846,\\\n84  0.466100 0.108322 0.429125,\\\n85  0.472328 0.110547 0.428334,\\\n86  0.478558 0.112764 0.427475,\\\n87  0.484789 0.114974 0.426548,\\\n88  0.491022 0.117179 0.425552,\\\n89  0.497257 0.119379 0.424488,\\\n90  0.503493 0.121575 0.423356,\\\n91  0.509730 0.123769 0.422156,\\\n92  0.515967 0.125960 0.420887,\\\n93  0.522206 0.128150 0.419549,\\\n94  0.528444 0.130341 0.418142,\\\n95  0.534683 0.132534 0.416667,\\\n96  0.540920 0.134729 0.415123,\\\n97  0.547157 0.136929 0.413511,\\\n98  0.553392 0.139134 0.411829,\\\n99  0.559624 0.141346 0.410078,\\\n100 0.565854 0.143567 0.408258,\\\n101 0.572081 0.145797 0.406369,\\\n102 0.578304 0.148039 0.404411,\\\n103 0.584521 0.150294 0.402385,\\\n104 0.590734 0.152563 0.400290,\\\n105 0.596940 0.154848 0.398125,\\\n106 0.603139 0.157151 0.395891,\\\n107 0.609330 0.159474 0.393589,\\\n108 0.615513 0.161817 0.391219,\\\n109 0.621685 0.164184 0.388781,\\\n110 0.627847 0.166575 0.386276,\\\n111 0.633998 0.168992 0.383704,\\\n112 0.640135 0.171438 0.381065,\\\n113 0.646260 0.173914 0.378359,\\\n114 0.652369 0.176421 0.375586,\\\n115 0.658463 0.178962 0.372748,\\\n116 0.664540 0.181539 0.369846,\\\n117 0.670599 0.184153 0.366879,\\\n118 0.676638 0.186807 0.363849,\\\n119 0.682656 0.189501 0.360757,\\\n120 0.688653 0.192239 0.357603,\\\n121 0.694627 0.195021 0.354388,\\\n122 0.700576 0.197851 0.351113,\\\n123 0.706500 0.200728 0.347777,\\\n124 0.712396 0.203656 0.344383,\\\n125 0.718264 0.206636 0.340931,\\\n126 0.724103 0.209670 0.337424,\\\n127 0.729909 0.212759 0.333861,\\\n128 0.735683 0.215906 0.330245,\\\n129 0.741423 0.219112 0.326576,\\\n130 0.747127 0.222378 0.322856,\\\n131 0.752794 0.225706 0.319085,\\\n132 0.758422 0.229097 0.315266,\\\n133 0.764010 0.232554 0.311399,\\\n134 0.769556 0.236077 0.307485,\\\n135 0.775059 0.239667 0.303526,\\\n136 0.780517 0.243327 0.299523,\\\n137 0.785929 0.247056 0.295477,\\\n138 0.791293 0.250856 0.291390,\\\n139 0.796607 0.254728 0.287264,\\\n140 0.801871 0.258674 0.283099,\\\n141 0.807082 0.262692 0.278898,\\\n142 0.812239 0.266786 0.274661,\\\n143 0.817341 0.270954 0.270390,\\\n144 0.822386 0.275197 0.266085,\\\n145 0.827372 0.279517 0.261750,\\\n146 0.832299 0.283913 0.257383,\\\n147 0.837165 0.288385 0.252988,\\\n148 0.841969 0.292933 0.248564,\\\n149 0.846709 0.297559 0.244113,\\\n150 0.851384 0.302260 0.239636,\\\n151 0.855992 0.307038 0.235133,\\\n152 0.860533 0.311892 0.230606,\\\n153 0.865006 0.316822 0.226055,\\\n154 0.869409 0.321827 0.221482,\\\n155 0.873741 0.326906 0.216886,\\\n156 0.878001 0.332060 0.212268,\\\n157 0.882188 0.337287 0.207628,\\\n158 0.886302 0.342586 0.202968,\\\n159 0.890341 0.347957 0.198286,\\\n160 0.894305 0.353399 0.193584,\\\n161 0.898192 0.358911 0.188860,\\\n162 0.902003 0.364492 0.184116,\\\n163 0.905735 0.370140 0.179350,\\\n164 0.909390 0.375856 0.174563,\\\n165 0.912966 0.381636 0.169755,\\\n166 0.916462 0.387481 0.164924,\\\n167 0.919879 0.393389 0.160070,\\\n168 0.923215 0.399359 0.155193,\\\n169 0.926470 0.405389 0.150292,\\\n170 0.929644 0.411479 0.145367,\\\n171 0.932737 0.417627 0.140417,\\\n172 0.935747 0.423831 0.135440,\\\n173 0.938675 0.430091 0.130438,\\\n174 0.941521 0.436405 0.125409,\\\n175 0.944285 0.442772 0.120354,\\\n176 0.946965 0.449191 0.115272,\\\n177 0.949562 0.455660 0.110164,\\\n178 0.952075 0.462178 0.105031,\\\n179 0.954506 0.468744 0.099874,\\\n180 0.956852 0.475356 0.094695,\\\n181 0.959114 0.482014 0.089499,\\\n182 0.961293 0.488716 0.084289,\\\n183 0.963387 0.495462 0.079073,\\\n184 0.965397 0.502249 0.073859,\\\n185 0.967322 0.509078 0.068659,\\\n186 0.969163 0.515946 0.063488,\\\n187 0.970919 0.522853 0.058367,\\\n188 0.972590 0.529798 0.053324,\\\n189 0.974176 0.536780 0.048392,\\\n190 0.975677 0.543798 0.043618,\\\n191 0.977092 0.550850 0.039050,\\\n192 0.978422 0.557937 0.034931,\\\n193 0.979666 0.565057 0.031409,\\\n194 0.980824 0.572209 0.028508,\\\n195 0.981895 0.579392 0.026250,\\\n196 0.982881 0.586606 0.024661,\\\n197 0.983779 0.593849 0.023770,\\\n198 0.984591 0.601122 0.023606,\\\n199 0.985315 0.608422 0.024202,\\\n200 0.985952 0.615750 0.025592,\\\n201 0.986502 0.623105 0.027814,\\\n202 0.986964 0.630485 0.030908,\\\n203 0.987337 0.637890 0.034916,\\\n204 0.987622 0.645320 0.039886,\\\n205 0.987819 0.652773 0.045581,\\\n206 0.987926 0.660250 0.051750,\\\n207 0.987945 0.667748 0.058329,\\\n208 0.987874 0.675267 0.065257,\\\n209 0.987714 0.682807 0.072489,\\\n210 0.987464 0.690366 0.079990,\\\n211 0.987124 0.697944 0.087731,\\\n212 0.986694 0.705540 0.095694,\\\n213 0.986175 0.713153 0.103863,\\\n214 0.985566 0.720782 0.112229,\\\n215 0.984865 0.728427 0.120785,\\\n216 0.984075 0.736087 0.129527,\\\n217 0.983196 0.743758 0.138453,\\\n218 0.982228 0.751442 0.147565,\\\n219 0.981173 0.759135 0.156863,\\\n220 0.980032 0.766837 0.166353,\\\n221 0.978806 0.774545 0.176037,\\\n222 0.977497 0.782258 0.185923,\\\n223 0.976108 0.789974 0.196018,\\\n224 0.974638 0.797692 0.206332,\\\n225 0.973088 0.805409 0.216877,\\\n226 0.971468 0.813122 0.227658,\\\n227 0.969783 0.820825 0.238686,\\\n228 0.968041 0.828515 0.249972,\\\n229 0.966243 0.836191 0.261534,\\\n230 0.964394 0.843848 0.273391,\\\n231 0.962517 0.851476 0.285546,\\\n232 0.960626 0.859069 0.298010,\\\n233 0.958720 0.866624 0.310820,\\\n234 0.956834 0.874129 0.323974,\\\n235 0.954997 0.881569 0.337475,\\\n236 0.953215 0.888942 0.351369,\\\n237 0.951546 0.896226 0.365627,\\\n238 0.950018 0.903409 0.380271,\\\n239 0.948683 0.910473 0.395289,\\\n240 0.947594 0.917399 0.410665,\\\n241 0.946809 0.924168 0.426373,\\\n242 0.946392 0.930761 0.442367,\\\n243 0.946403 0.937159 0.458592,\\\n244 0.946903 0.943348 0.474970,\\\n245 0.947937 0.949318 0.491426,\\\n246 0.949545 0.955063 0.507860,\\\n247 0.951740 0.960587 0.524203,\\\n248 0.954529 0.965896 0.540361,\\\n249 0.957896 0.971003 0.556275,\\\n250 0.961812 0.975924 0.571925,\\\n251 0.966249 0.980678 0.587206,\\\n252 0.971162 0.985282 0.602154,\\\n253 0.976511 0.989753 0.616760,\\\n254 0.982257 0.994109 0.631017,\\\n255 0.988362 0.998364 0.644924)\n" },
    { "jet", "# MATLAB jet color pallete\n\n# line styles\nset style line 1 lt 1 lc rgb '#000080' #\nset style line 2 lt 1 lc rgb '#0000ff' #\nset style line 3 lt 1 lc rgb '#0080ff' #\nset style line 4 lt 1 lc rgb '#00ffff' #\nset style line 5 lt 1 lc rgb '#80ff80' #\nset style line 6 lt 1 lc rgb '#ffff00' #\nset style line 7 lt 1 lc rgb '#ff8000' #\nset style line 8 lt 1 lc rgb '#ff0000' #\nset style line 9 lt 1 lc rgb '#800000' #\n# line style used together with jet (<2014b)\nset style line 11 lt 1 lc rgb '#0000ff' # blue\nset style line 12 lt 1 lc rgb '#007f00' # green\nset style line 13 lt 1 lc rgb '#ff0000' # red\nset style line 14 lt 1 lc rgb '#00bfbf' # cyan\nset style line 15 lt 1 lc rgb '#bf00bf' # pink\nset style line 16 lt 1 lc rgb '#bfbf00' # yellow\nset style line 17 lt 1 lc rgb '#3f3f3f' # black\n\n# palette\nset palette defined (0  0.0 0.0 0.5, \\\n                     1  0.0 0.0 1.0, \\\n                     2  0.0 0.5 1.0, \\\n                     3  0.0 1.0 1.0, \\\n                     4  0.5 1.0 0.5, \\\n                     5  1.0 1.0 0.0, \\\n                     6  1.0 0.5 0.0, \\\n                     7  1.0 0.0 0.0, \\\n                     8  0.5 0.0 0.0 )\n\n" },
    { "magma", "\n# New matplotlib colormaps by Nathaniel J. Smith, Stefan van der Walt,\n# and (in the case of viridis) Eric Firing.\n#\n# This file and the colormaps in it are released under the CC0 license /\n# public domain dedication. We would appreciate credit if you use or\n# redistribute these colormaps, but do not impose any legal restrictions.\n#\n# To the extent possible under law, the persons who associated CC0 with\n# mpl-colormaps have waived all copyright and related or neighboring rights\n# to mpl-colormaps.\n#\n# You should have received a copy of the CC0 legalcode along with this\n# work.  If not, see <http://creativecommons.org/publicdomain/zero/1.0/>.\n\n#https://github.com/BIDS/colormap/blob/master/colormaps.py\n\n\n# line styles\nset style line  1 lt 1 lc rgb '#000004' # black\nset style line  2 lt 1 lc rgb '#1c1044' # dark blue\nset style line  3 lt 1 lc rgb '#4f127b' # dark purple\nset style line  4 lt 1 lc rgb '#812581' # purple\nset style line  5 lt 1 lc rgb '#b5367a' # magenta\nset style line  6 lt 1 lc rgb '#e55964' # light red\nset style line  7 lt 1 lc rgb '#fb8761' # orange\nset style line  8 lt 1 lc rgb '#fec287' # light orange\nset style line  9 lt 1 lc rgb '#fbfdbf' # light yellow\n\n\n# palette\nset palette defined (\\\n0   0.001462 0.000466 0.013866,\\\n1   0.002258 0.001295 0.018331,\\\n2   0.003279 0.002305 0.023708,\\\n3   0.004512 0.003490 0.029965,\\\n4   0.005950 0.004843 0.037130,\\\n5   0.007588 0.006356 0.044973,\\\n6   0.009426 0.008022 0.052844,\\\n7   0.011465 0.009828 0.060750,\\\n8   0.013708 0.011771 0.068667,\\\n9   0.016156 0.013840 0.076603,\\\n10  0.018815 0.016026 0.084584,\\\n11  0.021692 0.018320 0.092610,\\\n12  0.024792 0.020715 0.100676,\\\n13  0.028123 0.023201 0.108787,\\\n14  0.031696 0.025765 0.116965,\\\n15  0.035520 0.028397 0.125209,\\\n16  0.039608 0.031090 0.133515,\\\n17  0.043830 0.033830 0.141886,\\\n18  0.048062 0.036607 0.150327,\\\n19  0.052320 0.039407 0.158841,\\\n20  0.056615 0.042160 0.167446,\\\n21  0.060949 0.044794 0.176129,\\\n22  0.065330 0.047318 0.184892,\\\n23  0.069764 0.049726 0.193735,\\\n24  0.074257 0.052017 0.202660,\\\n25  0.078815 0.054184 0.211667,\\\n26  0.083446 0.056225 0.220755,\\\n27  0.088155 0.058133 0.229922,\\\n28  0.092949 0.059904 0.239164,\\\n29  0.097833 0.061531 0.248477,\\\n30  0.102815 0.063010 0.257854,\\\n31  0.107899 0.064335 0.267289,\\\n32  0.113094 0.065492 0.276784,\\\n33  0.118405 0.066479 0.286321,\\\n34  0.123833 0.067295 0.295879,\\\n35  0.129380 0.067935 0.305443,\\\n36  0.135053 0.068391 0.315000,\\\n37  0.140858 0.068654 0.324538,\\\n38  0.146785 0.068738 0.334011,\\\n39  0.152839 0.068637 0.343404,\\\n40  0.159018 0.068354 0.352688,\\\n41  0.165308 0.067911 0.361816,\\\n42  0.171713 0.067305 0.370771,\\\n43  0.178212 0.066576 0.379497,\\\n44  0.184801 0.065732 0.387973,\\\n45  0.191460 0.064818 0.396152,\\\n46  0.198177 0.063862 0.404009,\\\n47  0.204935 0.062907 0.411514,\\\n48  0.211718 0.061992 0.418647,\\\n49  0.218512 0.061158 0.425392,\\\n50  0.225302 0.060445 0.431742,\\\n51  0.232077 0.059889 0.437695,\\\n52  0.238826 0.059517 0.443256,\\\n53  0.245543 0.059352 0.448436,\\\n54  0.252220 0.059415 0.453248,\\\n55  0.258857 0.059706 0.457710,\\\n56  0.265447 0.060237 0.461840,\\\n57  0.271994 0.060994 0.465660,\\\n58  0.278493 0.061978 0.469190,\\\n59  0.284951 0.063168 0.472451,\\\n60  0.291366 0.064553 0.475462,\\\n61  0.297740 0.066117 0.478243,\\\n62  0.304081 0.067835 0.480812,\\\n63  0.310382 0.069702 0.483186,\\\n64  0.316654 0.071690 0.485380,\\\n65  0.322899 0.073782 0.487408,\\\n66  0.329114 0.075972 0.489287,\\\n67  0.335308 0.078236 0.491024,\\\n68  0.341482 0.080564 0.492631,\\\n69  0.347636 0.082946 0.494121,\\\n70  0.353773 0.085373 0.495501,\\\n71  0.359898 0.087831 0.496778,\\\n72  0.366012 0.090314 0.497960,\\\n73  0.372116 0.092816 0.499053,\\\n74  0.378211 0.095332 0.500067,\\\n75  0.384299 0.097855 0.501002,\\\n76  0.390384 0.100379 0.501864,\\\n77  0.396467 0.102902 0.502658,\\\n78  0.402548 0.105420 0.503386,\\\n79  0.408629 0.107930 0.504052,\\\n80  0.414709 0.110431 0.504662,\\\n81  0.420791 0.112920 0.505215,\\\n82  0.426877 0.115395 0.505714,\\\n83  0.432967 0.117855 0.506160,\\\n84  0.439062 0.120298 0.506555,\\\n85  0.445163 0.122724 0.506901,\\\n86  0.451271 0.125132 0.507198,\\\n87  0.457386 0.127522 0.507448,\\\n88  0.463508 0.129893 0.507652,\\\n89  0.469640 0.132245 0.507809,\\\n90  0.475780 0.134577 0.507921,\\\n91  0.481929 0.136891 0.507989,\\\n92  0.488088 0.139186 0.508011,\\\n93  0.494258 0.141462 0.507988,\\\n94  0.500438 0.143719 0.507920,\\\n95  0.506629 0.145958 0.507806,\\\n96  0.512831 0.148179 0.507648,\\\n97  0.519045 0.150383 0.507443,\\\n98  0.525270 0.152569 0.507192,\\\n99  0.531507 0.154739 0.506895,\\\n100 0.537755 0.156894 0.506551,\\\n101 0.544015 0.159033 0.506159,\\\n102 0.550287 0.161158 0.505719,\\\n103 0.556571 0.163269 0.505230,\\\n104 0.562866 0.165368 0.504692,\\\n105 0.569172 0.167454 0.504105,\\\n106 0.575490 0.169530 0.503466,\\\n107 0.581819 0.171596 0.502777,\\\n108 0.588158 0.173652 0.502035,\\\n109 0.594508 0.175701 0.501241,\\\n110 0.600868 0.177743 0.500394,\\\n111 0.607238 0.179779 0.499492,\\\n112 0.613617 0.181811 0.498536,\\\n113 0.620005 0.183840 0.497524,\\\n114 0.626401 0.185867 0.496456,\\\n115 0.632805 0.187893 0.495332,\\\n116 0.639216 0.189921 0.494150,\\\n117 0.645633 0.191952 0.492910,\\\n118 0.652056 0.193986 0.491611,\\\n119 0.658483 0.196027 0.490253,\\\n120 0.664915 0.198075 0.488836,\\\n121 0.671349 0.200133 0.487358,\\\n122 0.677786 0.202203 0.485819,\\\n123 0.684224 0.204286 0.484219,\\\n124 0.690661 0.206384 0.482558,\\\n125 0.697098 0.208501 0.480835,\\\n126 0.703532 0.210638 0.479049,\\\n127 0.709962 0.212797 0.477201,\\\n128 0.716387 0.214982 0.475290,\\\n129 0.722805 0.217194 0.473316,\\\n130 0.729216 0.219437 0.471279,\\\n131 0.735616 0.221713 0.469180,\\\n132 0.742004 0.224025 0.467018,\\\n133 0.748378 0.226377 0.464794,\\\n134 0.754737 0.228772 0.462509,\\\n135 0.761077 0.231214 0.460162,\\\n136 0.767398 0.233705 0.457755,\\\n137 0.773695 0.236249 0.455289,\\\n138 0.779968 0.238851 0.452765,\\\n139 0.786212 0.241514 0.450184,\\\n140 0.792427 0.244242 0.447543,\\\n141 0.798608 0.247040 0.444848,\\\n142 0.804752 0.249911 0.442102,\\\n143 0.810855 0.252861 0.439305,\\\n144 0.816914 0.255895 0.436461,\\\n145 0.822926 0.259016 0.433573,\\\n146 0.828886 0.262229 0.430644,\\\n147 0.834791 0.265540 0.427671,\\\n148 0.840636 0.268953 0.424666,\\\n149 0.846416 0.272473 0.421631,\\\n150 0.852126 0.276106 0.418573,\\\n151 0.857763 0.279857 0.415496,\\\n152 0.863320 0.283729 0.412403,\\\n153 0.868793 0.287728 0.409303,\\\n154 0.874176 0.291859 0.406205,\\\n155 0.879464 0.296125 0.403118,\\\n156 0.884651 0.300530 0.400047,\\\n157 0.889731 0.305079 0.397002,\\\n158 0.894700 0.309773 0.393995,\\\n159 0.899552 0.314616 0.391037,\\\n160 0.904281 0.319610 0.388137,\\\n161 0.908884 0.324755 0.385308,\\\n162 0.913354 0.330052 0.382563,\\\n163 0.917689 0.335500 0.379915,\\\n164 0.921884 0.341098 0.377376,\\\n165 0.925937 0.346844 0.374959,\\\n166 0.929845 0.352734 0.372677,\\\n167 0.933606 0.358764 0.370541,\\\n168 0.937221 0.364929 0.368567,\\\n169 0.940687 0.371224 0.366762,\\\n170 0.944006 0.377643 0.365136,\\\n171 0.947180 0.384178 0.363701,\\\n172 0.950210 0.390820 0.362468,\\\n173 0.953099 0.397563 0.361438,\\\n174 0.955849 0.404400 0.360619,\\\n175 0.958464 0.411324 0.360014,\\\n176 0.960949 0.418323 0.359630,\\\n177 0.963310 0.425390 0.359469,\\\n178 0.965549 0.432519 0.359529,\\\n179 0.967671 0.439703 0.359810,\\\n180 0.969680 0.446936 0.360311,\\\n181 0.971582 0.454210 0.361030,\\\n182 0.973381 0.461520 0.361965,\\\n183 0.975082 0.468861 0.363111,\\\n184 0.976690 0.476226 0.364466,\\\n185 0.978210 0.483612 0.366025,\\\n186 0.979645 0.491014 0.367783,\\\n187 0.981000 0.498428 0.369734,\\\n188 0.982279 0.505851 0.371874,\\\n189 0.983485 0.513280 0.374198,\\\n190 0.984622 0.520713 0.376698,\\\n191 0.985693 0.528148 0.379371,\\\n192 0.986700 0.535582 0.382210,\\\n193 0.987646 0.543015 0.385210,\\\n194 0.988533 0.550446 0.388365,\\\n195 0.989363 0.557873 0.391671,\\\n196 0.990138 0.565296 0.395122,\\\n197 0.990871 0.572706 0.398714,\\\n198 0.991558 0.580107 0.402441,\\\n199 0.992196 0.587502 0.406299,\\\n200 0.992785 0.594891 0.410283,\\\n201 0.993326 0.602275 0.414390,\\\n202 0.993834 0.609644 0.418613,\\\n203 0.994309 0.616999 0.422950,\\\n204 0.994738 0.624350 0.427397,\\\n205 0.995122 0.631696 0.431951,\\\n206 0.995480 0.639027 0.436607,\\\n207 0.995810 0.646344 0.441361,\\\n208 0.996096 0.653659 0.446213,\\\n209 0.996341 0.660969 0.451160,\\\n210 0.996580 0.668256 0.456192,\\\n211 0.996775 0.675541 0.461314,\\\n212 0.996925 0.682828 0.466526,\\\n213 0.997077 0.690088 0.471811,\\\n214 0.997186 0.697349 0.477182,\\\n215 0.997254 0.704611 0.482635,\\\n216 0.997325 0.711848 0.488154,\\\n217 0.997351 0.719089 0.493755,\\\n218 0.997351 0.726324 0.499428,\\\n219 0.997341 0.733545 0.505167,\\\n220 0.997285 0.740772 0.510983,\\\n221 0.997228 0.747981 0.516859,\\\n222 0.997138 0.755190 0.522806,\\\n223 0.997019 0.762398 0.528821,\\\n224 0.996898 0.769591 0.534892,\\\n225 0.996727 0.776795 0.541039,\\\n226 0.996571 0.783977 0.547233,\\\n227 0.996369 0.791167 0.553499,\\\n228 0.996162 0.798348 0.559820,\\\n229 0.995932 0.805527 0.566202,\\\n230 0.995680 0.812706 0.572645,\\\n231 0.995424 0.819875 0.579140,\\\n232 0.995131 0.827052 0.585701,\\\n233 0.994851 0.834213 0.592307,\\\n234 0.994524 0.841387 0.598983,\\\n235 0.994222 0.848540 0.605696,\\\n236 0.993866 0.855711 0.612482,\\\n237 0.993545 0.862859 0.619299,\\\n238 0.993170 0.870024 0.626189,\\\n239 0.992831 0.877168 0.633109,\\\n240 0.992440 0.884330 0.640099,\\\n241 0.992089 0.891470 0.647116,\\\n242 0.991688 0.898627 0.654202,\\\n243 0.991332 0.905763 0.661309,\\\n244 0.990930 0.912915 0.668481,\\\n245 0.990570 0.920049 0.675675,\\\n246 0.990175 0.927196 0.682926,\\\n247 0.989815 0.934329 0.690198,\\\n248 0.989434 0.941470 0.697519,\\\n249 0.989077 0.948604 0.704863,\\\n250 0.988717 0.955742 0.712242,\\\n251 0.988367 0.962878 0.719649,\\\n252 0.988033 0.970012 0.727077,\\\n253 0.987691 0.977154 0.734536,\\\n254 0.987387 0.984288 0.742002,\\\n255 0.987053 0.991438 0.749504)\n" },
    { "matlab", "# Matlab color map parula, see:\n# http://www.mathworks.de/products/matlab/matlab-graphics/#new_look_for_matlab_graphics\n\n# New default Matlab line colors, introduced together with parula (2014b)\nset style line 1 lt 1 lc rgb '#0072bd' # blue\nset style line 2 lt 1 lc rgb '#d95319' # orange\nset style line 3 lt 1 lc rgb '#edb120' # yellow\nset style line 4 lt 1 lc rgb '#7e2f8e' # purple\nset style line 5 lt 1 lc rgb '#77ac30' # green\nset style line 6 lt 1 lc rgb '#4dbeee' # light-blue\nset style line 7 lt 1 lc rgb '#a2142f' # red\n\n# palette\nset palette defined (\\\n 0    0.2081    0.1663    0.5292,\\\n 1    0.2116    0.1898    0.5777,\\\n 2    0.2123    0.2138    0.6270,\\\n 3    0.2081    0.2386    0.6771,\\\n 4    0.1959    0.2645    0.7279,\\\n 5    0.1707    0.2919    0.7792,\\\n 6    0.1253    0.3242    0.8303,\\\n 7    0.0591    0.3598    0.8683,\\\n 8    0.0117    0.3875    0.8820,\\\n 9    0.0060    0.4086    0.8828,\\\n10    0.0165    0.4266    0.8786,\\\n11    0.0329    0.4430    0.8720,\\\n12    0.0498    0.4586    0.8641,\\\n13    0.0629    0.4737    0.8554,\\\n14    0.0723    0.4887    0.8467,\\\n15    0.0779    0.5040    0.8384,\\\n16    0.0793    0.5200    0.8312,\\\n17    0.0749    0.5375    0.8263,\\\n18    0.0641    0.5570    0.8240,\\\n19    0.0488    0.5772    0.8228,\\\n20    0.0343    0.5966    0.8199,\\\n21    0.0265    0.6137    0.8135,\\\n22    0.0239    0.6287    0.8038,\\\n23    0.0231    0.6418    0.7913,\\\n24    0.0228    0.6535    0.7768,\\\n25    0.0267    0.6642    0.7607,\\\n26    0.0384    0.6743    0.7436,\\\n27    0.0590    0.6838    0.7254,\\\n28    0.0843    0.6928    0.7062,\\\n29    0.1133    0.7015    0.6859,\\\n30    0.1453    0.7098    0.6646,\\\n31    0.1801    0.7177    0.6424,\\\n32    0.2178    0.7250    0.6193,\\\n33    0.2586    0.7317    0.5954,\\\n34    0.3022    0.7376    0.5712,\\\n35    0.3482    0.7424    0.5473,\\\n36    0.3953    0.7459    0.5244,\\\n37    0.4420    0.7481    0.5033,\\\n38    0.4871    0.7491    0.4840,\\\n39    0.5300    0.7491    0.4661,\\\n40    0.5709    0.7485    0.4494,\\\n41    0.6099    0.7473    0.4337,\\\n42    0.6473    0.7456    0.4188,\\\n43    0.6834    0.7435    0.4044,\\\n44    0.7184    0.7411    0.3905,\\\n45    0.7525    0.7384    0.3768,\\\n46    0.7858    0.7356    0.3633,\\\n47    0.8185    0.7327    0.3498,\\\n48    0.8507    0.7299    0.3360,\\\n49    0.8824    0.7274    0.3217,\\\n50    0.9139    0.7258    0.3063,\\\n51    0.9450    0.7261    0.2886,\\\n52    0.9739    0.7314    0.2666,\\\n53    0.9938    0.7455    0.2403,\\\n54    0.9990    0.7653    0.2164,\\\n55    0.9955    0.7861    0.1967,\\\n56    0.9880    0.8066    0.1794,\\\n57    0.9789    0.8271    0.1633,\\\n58    0.9697    0.8481    0.1475,\\\n59    0.9626    0.8705    0.1309,\\\n60    0.9589    0.8949    0.1132,\\\n61    0.9598    0.9218    0.0948,\\\n62    0.9661    0.9514    0.0755,\\\n63    0.9763    0.9831    0.0538)" },
    { "moreland", "# Blue/Red color pallete after Moreland (2009)\n# http://www.sandia.gov/~kmorel/documents/ColorMaps/\n# http://bastian.rieck.ru/blog/posts/2012/gnuplot_better_colour_palettes/\n\n# line styles\nset style line 1 lt 1 lc rgb '#3b4cc0' #\nset style line 2 lt 1 lc rgb '#688aef' #\nset style line 3 lt 1 lc rgb '#99baff' #\nset style line 4 lt 1 lc rgb '#c9d8ef' #\nset style line 5 lt 1 lc rgb '#edd1c2' #\nset style line 6 lt 1 lc rgb '#f7a789' #\nset style line 7 lt 1 lc rgb '#e36a53' #\nset style line 8 lt 1 lc rgb '#b40426' #\n\n# palette\n# TODO: check if the short version gives the same results\nset palette defined (\\\n0 0.2298057 0.298717966 0.753683153,\\\n0.00390625 0.234299935 0.305559204 0.759874796,\\\n0.0078125 0.238810063 0.312388385 0.766005866,\\\n0.01171875 0.243336663 0.319205292 0.772075394,\\\n0.015625 0.247880265 0.326009656 0.778082421,\\\n0.01953125 0.25244136 0.332801165 0.784026001,\\\n0.0234375 0.257020396 0.339579464 0.789905199,\\\n0.02734375 0.261617779 0.346344164 0.79571909,\\\n0.03125 0.26623388 0.353094838 0.801466763,\\\n0.03515625 0.270869029 0.359831032 0.807147315,\\\n0.0390625 0.275523523 0.36655226 0.812759858,\\\n0.04296875 0.28019762 0.373258014 0.818303516,\\\n0.046875 0.284891546 0.379947761 0.823777422,\\\n0.05078125 0.289605495 0.386620945 0.829180725,\\\n0.0546875 0.294339624 0.393276993 0.834512584,\\\n0.05859375 0.299094064 0.399915313 0.839772171,\\\n0.0625 0.30386891 0.406535296 0.84495867,\\\n0.06640625 0.308664231 0.413136319 0.850071279,\\\n0.0703125 0.313480065 0.419717745 0.855109207,\\\n0.07421875 0.318316422 0.426278924 0.860071679,\\\n0.078125 0.323173283 0.432819194 0.864957929,\\\n0.08203125 0.328050603 0.439337884 0.869767207,\\\n0.0859375 0.332948312 0.445834313 0.874498775,\\\n0.08984375 0.337866311 0.45230779 0.87915191,\\\n0.09375 0.342804478 0.458757618 0.883725899,\\\n0.09765625 0.347762667 0.465183092 0.888220047,\\\n0.1015625 0.352740705 0.471583499 0.892633669,\\\n0.10546875 0.357738399 0.477958123 0.896966095,\\\n0.109375 0.362755532 0.484306241 0.90121667,\\\n0.11328125 0.367791863 0.490627125 0.905384751,\\\n0.1171875 0.372847134 0.496920043 0.909469711,\\\n0.12109375 0.37792106 0.503184261 0.913470934,\\\n0.125 0.38301334 0.50941904 0.917387822,\\\n0.12890625 0.38812365 0.515623638 0.921219788,\\\n0.1328125 0.39325165 0.521797312 0.924966262,\\\n0.13671875 0.398396976 0.527939316 0.928626686,\\\n0.140625 0.40355925 0.534048902 0.932200518,\\\n0.14453125 0.408738074 0.540125323 0.93568723,\\\n0.1484375 0.413933033 0.546167829 0.939086309,\\\n0.15234375 0.419143694 0.552175668 0.942397257,\\\n0.15625 0.424369608 0.558148092 0.945619588,\\\n0.16015625 0.429610311 0.564084349 0.948752835,\\\n0.1640625 0.434865321 0.56998369 0.951796543,\\\n0.16796875 0.440134144 0.575845364 0.954750272,\\\n0.171875 0.445416268 0.581668623 0.957613599,\\\n0.17578125 0.450711169 0.587452719 0.960386113,\\\n0.1796875 0.456018308 0.593196905 0.96306742,\\\n0.18359375 0.461337134 0.598900436 0.96565714,\\\n0.1875 0.46666708 0.604562568 0.968154911,\\\n0.19140625 0.472007569 0.61018256 0.970560381,\\\n0.1953125 0.477358011 0.615759672 0.972873218,\\\n0.19921875 0.482717804 0.621293167 0.975093102,\\\n0.203125 0.488086336 0.626782311 0.97721973,\\\n0.20703125 0.493462982 0.632226371 0.979252813,\\\n0.2109375 0.498847107 0.637624618 0.981192078,\\\n0.21484375 0.504238066 0.642976326 0.983037268,\\\n0.21875 0.509635204 0.648280772 0.98478814,\\\n0.22265625 0.515037856 0.653537236 0.986444467,\\\n0.2265625 0.520445349 0.658745003 0.988006036,\\\n0.23046875 0.525857 0.66390336 0.989472652,\\\n0.234375 0.531272118 0.669011598 0.990844132,\\\n0.23828125 0.536690004 0.674069012 0.99212031,\\\n0.2421875 0.542109949 0.679074903 0.993301037,\\\n0.24609375 0.54753124 0.684028574 0.994386177,\\\n0.25 0.552953156 0.688929332 0.995375608,\\\n0.25390625 0.558374965 0.693776492 0.996269227,\\\n0.2578125 0.563795935 0.698569369 0.997066945,\\\n0.26171875 0.569215322 0.703307287 0.997768685,\\\n0.265625 0.574632379 0.707989572 0.99837439,\\\n0.26953125 0.580046354 0.712615557 0.998884016,\\\n0.2734375 0.585456486 0.717184578 0.999297533,\\\n0.27734375 0.590862011 0.721695979 0.999614929,\\\n0.28125 0.596262162 0.726149107 0.999836203,\\\n0.28515625 0.601656165 0.730543315 0.999961374,\\\n0.2890625 0.607043242 0.734877964 0.999990472,\\\n0.29296875 0.61242261 0.739152418 0.999923544,\\\n0.296875 0.617793485 0.743366047 0.999760652,\\\n0.30078125 0.623155076 0.747518228 0.999501871,\\\n0.3046875 0.628506592 0.751608345 0.999147293,\\\n0.30859375 0.633847237 0.755635786 0.998697024,\\\n0.3125 0.639176211 0.759599947 0.998151185,\\\n0.31640625 0.644492714 0.763500228 0.99750991,\\\n0.3203125 0.649795942 0.767336039 0.996773351,\\\n0.32421875 0.655085089 0.771106793 0.995941671,\\\n0.328125 0.660359348 0.774811913 0.995015049,\\\n0.33203125 0.665617908 0.778450826 0.993993679,\\\n0.3359375 0.670859959 0.782022968 0.992877768,\\\n0.33984375 0.676084688 0.78552778 0.991667539,\\\n0.34375 0.681291281 0.788964712 0.990363227,\\\n0.34765625 0.686478925 0.792333219 0.988965083,\\\n0.3515625 0.691646803 0.795632765 0.987473371,\\\n0.35546875 0.696794099 0.798862821 0.985888369,\\\n0.359375 0.701919999 0.802022864 0.984210369,\\\n0.36328125 0.707023684 0.805112381 0.982439677,\\\n0.3671875 0.712104339 0.808130864 0.980576612,\\\n0.37109375 0.717161148 0.811077814 0.978621507,\\\n0.375 0.722193294 0.813952739 0.976574709,\\\n0.37890625 0.727199962 0.816755156 0.974436577,\\\n0.3828125 0.732180337 0.81948459 0.972207484,\\\n0.38671875 0.737133606 0.82214057 0.969887816,\\\n0.390625 0.742058956 0.824722639 0.967477972,\\\n0.39453125 0.746955574 0.827230344 0.964978364,\\\n0.3984375 0.751822652 0.829663241 0.962389418,\\\n0.40234375 0.756659379 0.832020895 0.959711569,\\\n0.40625 0.761464949 0.834302879 0.956945269,\\\n0.41015625 0.766238556 0.836508774 0.95409098,\\\n0.4140625 0.770979397 0.838638169 0.951149176,\\\n0.41796875 0.775686671 0.840690662 0.948120345,\\\n0.421875 0.780359577 0.842665861 0.945004985,\\\n0.42578125 0.78499732 0.84456338 0.941803607,\\\n0.4296875 0.789599105 0.846382843 0.938516733,\\\n0.43359375 0.79416414 0.848123884 0.935144898,\\\n0.4375 0.798691636 0.849786142 0.931688648,\\\n0.44140625 0.803180808 0.85136927 0.928148539,\\\n0.4453125 0.807630872 0.852872925 0.92452514,\\\n0.44921875 0.812041048 0.854296776 0.92081903,\\\n0.453125 0.81641056 0.855640499 0.917030798,\\\n0.45703125 0.820738635 0.856903782 0.913161047,\\\n0.4609375 0.825024503 0.85808632 0.909210387,\\\n0.46484375 0.829267397 0.859187816 0.90517944,\\\n0.46875 0.833466556 0.860207984 0.901068838,\\\n0.47265625 0.837621221 0.861146547 0.896879224,\\\n0.4765625 0.841730637 0.862003236 0.892611249,\\\n0.48046875 0.845794055 0.862777795 0.888265576,\\\n0.484375 0.849810727 0.863469972 0.883842876,\\\n0.48828125 0.853779913 0.864079527 0.87934383,\\\n0.4921875 0.857700874 0.864606232 0.874769128,\\\n0.49609375 0.861572878 0.865049863 0.870119469,\\\n0.5 0.865395197 0.86541021 0.865395561,\\\n0.50390625 0.86977749 0.863633958 0.859948576,\\\n0.5078125 0.874064226 0.861776352 0.854466231,\\\n0.51171875 0.878255583 0.859837644 0.848949435,\\\n0.515625 0.882351728 0.857818097 0.843399101,\\\n0.51953125 0.886352818 0.85571798 0.837816138,\\\n0.5234375 0.890259 0.853537573 0.832201453,\\\n0.52734375 0.89407041 0.851277164 0.826555954,\\\n0.53125 0.897787179 0.848937047 0.820880546,\\\n0.53515625 0.901409427 0.846517528 0.815176131,\\\n0.5390625 0.904937269 0.844018919 0.809443611,\\\n0.54296875 0.908370816 0.841441541 0.803683885,\\\n0.546875 0.911710171 0.838785722 0.79789785,\\\n0.55078125 0.914955433 0.836051799 0.792086401,\\\n0.5546875 0.918106696 0.833240115 0.786250429,\\\n0.55859375 0.921164054 0.830351023 0.780390824,\\\n0.5625 0.924127593 0.827384882 0.774508472,\\\n0.56640625 0.926997401 0.824342058 0.768604257,\\\n0.5703125 0.929773562 0.821222926 0.76267906,\\\n0.57421875 0.932456159 0.818027865 0.756733758,\\\n0.578125 0.935045272 0.814757264 0.750769226,\\\n0.58203125 0.937540984 0.811411517 0.744786333,\\\n0.5859375 0.939943375 0.807991025 0.738785947,\\\n0.58984375 0.942252526 0.804496196 0.732768931,\\\n0.59375 0.944468518 0.800927443 0.726736146,\\\n0.59765625 0.946591434 0.797285187 0.720688446,\\\n0.6015625 0.948621357 0.793569853 0.714626683,\\\n0.60546875 0.950558373 0.789781872 0.708551706,\\\n0.609375 0.952402567 0.785921682 0.702464356,\\\n0.61328125 0.954154029 0.781989725 0.696365473,\\\n0.6171875 0.955812849 0.777986449 0.690255891,\\\n0.62109375 0.957379123 0.773912305 0.68413644,\\\n0.625 0.958852946 0.769767752 0.678007945,\\\n0.62890625 0.960234418 0.765553251 0.671871226,\\\n0.6328125 0.961523642 0.761269267 0.665727098,\\\n0.63671875 0.962720725 0.756916272 0.659576372,\\\n0.640625 0.963825777 0.752494738 0.653419853,\\\n0.64453125 0.964838913 0.748005143 0.647258341,\\\n0.6484375 0.965760251 0.743447967 0.64109263,\\\n0.65234375 0.966589914 0.738823693 0.634923509,\\\n0.65625 0.96732803 0.734132809 0.628751763,\\\n0.66015625 0.967974729 0.729375802 0.62257817,\\\n0.6640625 0.96853015 0.724553162 0.616403502,\\\n0.66796875 0.968994435 0.719665383 0.610228525,\\\n0.671875 0.969367729 0.714712956 0.604054002,\\\n0.67578125 0.969650186 0.709696378 0.597880686,\\\n0.6796875 0.969841963 0.704616143 0.591709328,\\\n0.68359375 0.969943224 0.699472746 0.585540669,\\\n0.6875 0.969954137 0.694266682 0.579375448,\\\n0.69140625 0.969874878 0.688998447 0.573214394,\\\n0.6953125 0.969705626 0.683668532 0.567058232,\\\n0.69921875 0.96944657 0.678277431 0.560907681,\\\n0.703125 0.969097901 0.672825633 0.554763452,\\\n0.70703125 0.968659818 0.667313624 0.54862625,\\\n0.7109375 0.968132528 0.661741889 0.542496774,\\\n0.71484375 0.967516241 0.656110908 0.536375716,\\\n0.71875 0.966811177 0.650421156 0.530263762,\\\n0.72265625 0.966017559 0.644673104 0.524161591,\\\n0.7265625 0.965135621 0.638867216 0.518069875,\\\n0.73046875 0.964165599 0.63300395 0.511989279,\\\n0.734375 0.963107739 0.627083758 0.505920462,\\\n0.73828125 0.961962293 0.621107082 0.499864075,\\\n0.7421875 0.960729521 0.615074355 0.493820764,\\\n0.74609375 0.959409687 0.608986 0.487791167,\\\n0.75 0.958003065 0.602842431 0.481775914,\\\n0.75390625 0.956509936 0.596644046 0.475775629,\\\n0.7578125 0.954930586 0.590391232 0.46979093,\\\n0.76171875 0.95326531 0.584084361 0.463822426,\\\n0.765625 0.951514411 0.57772379 0.457870719,\\\n0.76953125 0.949678196 0.571309856 0.451936407,\\\n0.7734375 0.947756983 0.564842879 0.446020077,\\\n0.77734375 0.945751096 0.558323158 0.440122312,\\\n0.78125 0.943660866 0.551750968 0.434243684,\\\n0.78515625 0.941486631 0.545126562 0.428384763,\\\n0.7890625 0.939228739 0.538450165 0.422546107,\\\n0.79296875 0.936887543 0.531721972 0.41672827,\\\n0.796875 0.934463404 0.524942147 0.410931798,\\\n0.80078125 0.931956691 0.518110821 0.40515723,\\\n0.8046875 0.929367782 0.511228087 0.399405096,\\\n0.80859375 0.92669706 0.504293997 0.393675922,\\\n0.8125 0.923944917 0.49730856 0.387970225,\\\n0.81640625 0.921111753 0.490271735 0.382288516,\\\n0.8203125 0.918197974 0.483183431 0.376631297,\\\n0.82421875 0.915203996 0.476043498 0.370999065,\\\n0.828125 0.912130241 0.468851724 0.36539231,\\\n0.83203125 0.908977139 0.461607831 0.359811513,\\\n0.8359375 0.905745128 0.454311462 0.354257151,\\\n0.83984375 0.902434654 0.446962183 0.348729691,\\\n0.84375 0.89904617 0.439559467 0.343229596,\\\n0.84765625 0.895580136 0.43210269 0.33775732,\\\n0.8515625 0.892037022 0.424591118 0.332313313,\\\n0.85546875 0.888417303 0.417023898 0.326898016,\\\n0.859375 0.884721464 0.409400045 0.321511863,\\\n0.86328125 0.880949996 0.401718425 0.316155284,\\\n0.8671875 0.877103399 0.393977745 0.310828702,\\\n0.87109375 0.873182178 0.386176527 0.305532531,\\\n0.875 0.869186849 0.378313092 0.300267182,\\\n0.87890625 0.865117934 0.370385535 0.295033059,\\\n0.8828125 0.860975962 0.362391695 0.289830559,\\\n0.88671875 0.85676147 0.354329127 0.284660075,\\\n0.890625 0.852475004 0.346195061 0.279521991,\\\n0.89453125 0.848117114 0.337986361 0.27441669,\\\n0.8984375 0.843688361 0.329699471 0.269344545,\\\n0.90234375 0.839189312 0.32133036 0.264305927,\\\n0.90625 0.834620542 0.312874446 0.259301199,\\\n0.91015625 0.829982631 0.304326513 0.254330723,\\\n0.9140625 0.82527617 0.295680611 0.249394851,\\\n0.91796875 0.820501754 0.286929926 0.244493934,\\\n0.921875 0.815659988 0.278066636 0.239628318,\\\n0.92578125 0.810751482 0.269081721 0.234798343,\\\n0.9296875 0.805776855 0.259964733 0.230004348,\\\n0.93359375 0.800736732 0.250703507 0.225246666,\\\n0.9375 0.795631745 0.24128379 0.220525627,\\\n0.94140625 0.790462533 0.231688768 0.215841558,\\\n0.9453125 0.785229744 0.221898442 0.211194782,\\\n0.94921875 0.779934029 0.211888813 0.20658562,\\\n0.953125 0.774576051 0.201630762 0.202014392,\\\n0.95703125 0.769156474 0.191088518 0.197481414,\\\n0.9609375 0.763675975 0.180217488 0.192987001,\\\n0.96484375 0.758135232 0.168961101 0.188531467,\\\n0.96875 0.752534934 0.157246067 0.184115123,\\\n0.97265625 0.746875773 0.144974956 0.179738284,\\\n0.9765625 0.741158452 0.132014017 0.175401259,\\\n0.98046875 0.735383675 0.1181719 0.171104363,\\\n0.984375 0.729552157 0.103159409 0.166847907,\\\n0.98828125 0.723664618 0.086504694 0.162632207,\\\n0.9921875 0.717721782 0.067344036 0.158457578,\\\n0.99609375 0.711724383 0.043755173 0.154324339,\\\n1 0.705673158 0.01555616 0.150232812)\n" },
    { "oranges", "# line styles for ColorBrewer Oranges\n# for use with sequential data\n# provides 8 orange colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#FFF5EB' # very light orange\nset style line 2 lt 1 lc rgb '#FEE6CE' # \nset style line 3 lt 1 lc rgb '#FDD0A2' # \nset style line 4 lt 1 lc rgb '#FDAE6B' # light orange\nset style line 5 lt 1 lc rgb '#FD8D3C' # \nset style line 6 lt 1 lc rgb '#F16913' # medium orange\nset style line 7 lt 1 lc rgb '#D94801' #\nset style line 8 lt 1 lc rgb '#8C2D04' # dark orange\n\n# palette\nset palette defined ( 0 '#FFF5EB',\\\n    \t    \t      1 '#FEE6CE',\\\n\t\t      2 '#FDD0A2',\\\n\t\t      3 '#FDAE6B',\\\n\t\t      4 '#FD8D3C',\\\n\t\t      5 '#F16913',\\\n\t\t      6 '#D94801',\\\n\t\t      7 '#8C2D04' )\n" },
    { "orrd", "# line styles for ColorBrewer OrRd\n# for use with sequential data\n# provides 8 orange-red colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#FFF7EC' # very light orange-red\nset style line 2 lt 1 lc rgb '#FEE8C8' # \nset style line 3 lt 1 lc rgb '#FDD49E' # \nset style line 4 lt 1 lc rgb '#FDBB84' # light orange-red\nset style line 5 lt 1 lc rgb '#FC8D59' # \nset style line 6 lt 1 lc rgb '#EF6548' # medium orange-red\nset style line 7 lt 1 lc rgb '#D7301F' #\nset style line 8 lt 1 lc rgb '#990000' # dark orange-red\n\n# palette\nset palette defined ( 0 '#FFF7EC',\\\n    \t    \t      1 '#FEE8C8',\\\n\t\t      2 '#FDD49E',\\\n\t\t      3 '#FDBB84',\\\n\t\t      4 '#FC8D59',\\\n\t\t      5 '#EF6548',\\\n\t\t      6 '#D7301F',\\\n\t\t      7 '#990000' )\n" },
    { "paired", "# line styles for ColorBrewer Paired\n# for use with qualitative/categorical data\n# provides 8 colors in 4 pairs\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#A6CEE3' # light blue\nset style line 2 lt 1 lc rgb '#1F78B4' # dark blue\nset style line 3 lt 1 lc rgb '#B2DF8A' # light green\nset style line 4 lt 1 lc rgb '#33A02C' # dark green\nset style line 5 lt 1 lc rgb '#FB9A99' # light red\nset style line 6 lt 1 lc rgb '#E31A1C' # dark red\nset style line 7 lt 1 lc rgb '#FDBF6F' # light orange\nset style line 8 lt 1 lc rgb '#FF7F00' # dark orange\n\n# palette\nset palette maxcolors 8\nset palette defined ( 0 '#A6CEE3',\\\n    \t    \t      1 '#1F78B4',\\\n\t\t      2 '#B2DF8A',\\\n\t\t      3 '#33A02C',\\\n\t\t      4 '#FB9A99',\\\n\t\t      5 '#E31A1C',\\\n\t\t      6 '#FDBF6F',\\\n\t\t      7 '#FF7F00' )\n" },
    { "parula", "# Colormap similar to Matlab parula, see:\n# http://www.mathworks.de/products/matlab/matlab-graphics/#new_look_for_matlab_graphics\n\n# line styles\nset style line  1 lt 1 lc rgb '#352a87' # blue\nset style line  2 lt 1 lc rgb '#0f5cdd' # blue\nset style line  3 lt 1 lc rgb '#1481d6' # blue\nset style line  4 lt 1 lc rgb '#06a4ca' # cyan\nset style line  5 lt 1 lc rgb '#2eb7a4' # green\nset style line  6 lt 1 lc rgb '#87bf77' # green\nset style line  7 lt 1 lc rgb '#d1bb59' # orange\nset style line  8 lt 1 lc rgb '#fec832' # orange\nset style line  9 lt 1 lc rgb '#f9fb0e' # yellow\n\n# New default Matlab line colors, introduced together with parula (2014b)\nset style line 11 lt 1 lc rgb '#0072bd' # blue\nset style line 12 lt 1 lc rgb '#d95319' # orange\nset style line 13 lt 1 lc rgb '#edb120' # yellow\nset style line 14 lt 1 lc rgb '#7e2f8e' # purple\nset style line 15 lt 1 lc rgb '#77ac30' # green\nset style line 16 lt 1 lc rgb '#4dbeee' # light-blue\nset style line 17 lt 1 lc rgb '#a2142f' # red\n\n# palette\nset palette defined (\\\n0 '#352a87',\\\n1 '#0363e1',\\\n2 '#1485d4',\\\n3 '#06a7c6',\\\n4 '#38b99e',\\\n5 '#92bf73',\\\n6 '#d9ba56',\\\n7 '#fcce2e',\\\n8 '#f9fb0e')\n" },
    { "pastel1", "# line styles for ColorBrewer Pastel1\n# for use with qualitative/categorical data\n# provides 8 pale colors based on Set1\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#FBB4AE' # pale red\nset style line 2 lt 1 lc rgb '#B3CDE3' # pale blue\nset style line 3 lt 1 lc rgb '#CCEBC5' # pale green\nset style line 4 lt 1 lc rgb '#DECBE4' # pale purple\nset style line 5 lt 1 lc rgb '#FED9A6' # pale orange\nset style line 6 lt 1 lc rgb '#FFFFCC' # pale yellow\nset style line 7 lt 1 lc rgb '#E5D8BD' # pale brown\nset style line 8 lt 1 lc rgb '#FDDAEC' # pale pink\n\n# palette\nset palette maxcolors 8\nset palette defined ( 0 '#FBB4AE',\\\n    \t    \t      1 '#B3CDE3',\\\n\t\t      2 '#CCEBC5',\\\n\t\t      3 '#DECBE4',\\\n\t\t      4 '#FED9A6',\\\n\t\t      5 '#FFFFCC',\\\n\t\t      6 '#E5D8BD',\\\n\t\t      7 '#FDDAEC' )\n" },
    { "pastel2", "# line styles for ColorBrewer Pastel2\n# for use with qualitative/categorical data\n# provides 8 pale colors based on Set2\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#B3E2CD' # pale teal\nset style line 2 lt 1 lc rgb '#FDCDAC' # pale orange\nset style line 3 lt 1 lc rgb '#CDB5E8' # pale lilac\nset style line 4 lt 1 lc rgb '#F4CAE4' # pale magenta\nset style line 5 lt 1 lc rgb '#D6F5C9' # pale lime green\nset style line 6 lt 1 lc rgb '#FFF2AE' # pale banana\nset style line 7 lt 1 lc rgb '#F1E2CC' # pale tan\nset style line 8 lt 1 lc rgb '#CCCCCC' # pale grey\n\n# palette\nset palette maxcolors 8\nset palette defined ( 0 '#B3E2CD',\\\n    \t    \t      1 '#FDCDAC',\\\n\t\t      2 '#CDB5E8',\\\n\t\t      3 '#F4CAE4',\\\n\t\t      4 '#D6F5C9',\\\n\t\t      5 '#FFF2AE',\\\n\t\t      6 '#F1E2CC',\\\n\t\t      7 '#CCCCCC' )\n" },
    { "piyg", "# line styles for ColorBrewer PiYG\n# for use with divering data\n# provides 8 colors with pink low, white middle, and yellow-green high\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#C51B7D' # dark pink\nset style line 2 lt 1 lc rgb '#DE77AE' # medium pink\nset style line 3 lt 1 lc rgb '#F1B6DA' # \nset style line 4 lt 1 lc rgb '#FDE0EF' # pale pink\nset style line 5 lt 1 lc rgb '#E6F5D0' # pale yellow-green\nset style line 6 lt 1 lc rgb '#B8E186' # \nset style line 7 lt 1 lc rgb '#7FBC41' # medium yellow-green\nset style line 8 lt 1 lc rgb '#4D9221' # dark yellow-green\n\n# palette\nset palette defined ( 0 '#C51B7D',\\\n    \t    \t      1 '#DE77AE',\\\n\t\t      2 '#F1B6DA',\\\n\t\t      3 '#FDE0EF',\\\n\t\t      4 '#E6F5D0',\\\n\t\t      5 '#B8E186',\\\n\t\t      6 '#7FBC41',\\\n\t\t      7 '#4D9221' )\n" },
    { "plasma", "# New matplotlib colormaps by Nathaniel J. Smith, Stefan van der Walt,\n# and (in the case of viridis) Eric Firing.\n#\n# This file and the colormaps in it are released under the CC0 license /\n# public domain dedication. We would appreciate credit if you use or\n# redistribute these colormaps, but do not impose any legal restrictions.\n#\n# To the extent possible under law, the persons who associated CC0 with\n# mpl-colormaps have waived all copyright and related or neighboring rights\n# to mpl-colormaps.\n#\n# You should have received a copy of the CC0 legalcode along with this\n# work.  If not, see <http://creativecommons.org/publicdomain/zero/1.0/>.\n\n#https://github.com/BIDS/colormap/blob/master/colormaps.py\n\n\n# line styles\nset style line  1 lt 1 lc rgb '#0c0887' # blue\nset style line  2 lt 1 lc rgb '#4b03a1' # purple-blue\nset style line  3 lt 1 lc rgb '#7d03a8' # purple\nset style line  4 lt 1 lc rgb '#a82296' # purple\nset style line  5 lt 1 lc rgb '#cb4679' # magenta\nset style line  6 lt 1 lc rgb '#e56b5d' # red\nset style line  7 lt 1 lc rgb '#f89441' # orange\nset style line  8 lt 1 lc rgb '#fdc328' # orange\nset style line  9 lt 1 lc rgb '#f0f921' # yellow\n\n\n# palette\nset palette defined (\\\n0   0.050383 0.029803 0.527975,\\\n1   0.063536 0.028426 0.533124,\\\n2   0.075353 0.027206 0.538007,\\\n3   0.086222 0.026125 0.542658,\\\n4   0.096379 0.025165 0.547103,\\\n5   0.105980 0.024309 0.551368,\\\n6   0.115124 0.023556 0.555468,\\\n7   0.123903 0.022878 0.559423,\\\n8   0.132381 0.022258 0.563250,\\\n9   0.140603 0.021687 0.566959,\\\n10  0.148607 0.021154 0.570562,\\\n11  0.156421 0.020651 0.574065,\\\n12  0.164070 0.020171 0.577478,\\\n13  0.171574 0.019706 0.580806,\\\n14  0.178950 0.019252 0.584054,\\\n15  0.186213 0.018803 0.587228,\\\n16  0.193374 0.018354 0.590330,\\\n17  0.200445 0.017902 0.593364,\\\n18  0.207435 0.017442 0.596333,\\\n19  0.214350 0.016973 0.599239,\\\n20  0.221197 0.016497 0.602083,\\\n21  0.227983 0.016007 0.604867,\\\n22  0.234715 0.015502 0.607592,\\\n23  0.241396 0.014979 0.610259,\\\n24  0.248032 0.014439 0.612868,\\\n25  0.254627 0.013882 0.615419,\\\n26  0.261183 0.013308 0.617911,\\\n27  0.267703 0.012716 0.620346,\\\n28  0.274191 0.012109 0.622722,\\\n29  0.280648 0.011488 0.625038,\\\n30  0.287076 0.010855 0.627295,\\\n31  0.293478 0.010213 0.629490,\\\n32  0.299855 0.009561 0.631624,\\\n33  0.306210 0.008902 0.633694,\\\n34  0.312543 0.008239 0.635700,\\\n35  0.318856 0.007576 0.637640,\\\n36  0.325150 0.006915 0.639512,\\\n37  0.331426 0.006261 0.641316,\\\n38  0.337683 0.005618 0.643049,\\\n39  0.343925 0.004991 0.644710,\\\n40  0.350150 0.004382 0.646298,\\\n41  0.356359 0.003798 0.647810,\\\n42  0.362553 0.003243 0.649245,\\\n43  0.368733 0.002724 0.650601,\\\n44  0.374897 0.002245 0.651876,\\\n45  0.381047 0.001814 0.653068,\\\n46  0.387183 0.001434 0.654177,\\\n47  0.393304 0.001114 0.655199,\\\n48  0.399411 0.000859 0.656133,\\\n49  0.405503 0.000678 0.656977,\\\n50  0.411580 0.000577 0.657730,\\\n51  0.417642 0.000564 0.658390,\\\n52  0.423689 0.000646 0.658956,\\\n53  0.429719 0.000831 0.659425,\\\n54  0.435734 0.001127 0.659797,\\\n55  0.441732 0.001540 0.660069,\\\n56  0.447714 0.002080 0.660240,\\\n57  0.453677 0.002755 0.660310,\\\n58  0.459623 0.003574 0.660277,\\\n59  0.465550 0.004545 0.660139,\\\n60  0.471457 0.005678 0.659897,\\\n61  0.477344 0.006980 0.659549,\\\n62  0.483210 0.008460 0.659095,\\\n63  0.489055 0.010127 0.658534,\\\n64  0.494877 0.011990 0.657865,\\\n65  0.500678 0.014055 0.657088,\\\n66  0.506454 0.016333 0.656202,\\\n67  0.512206 0.018833 0.655209,\\\n68  0.517933 0.021563 0.654109,\\\n69  0.523633 0.024532 0.652901,\\\n70  0.529306 0.027747 0.651586,\\\n71  0.534952 0.031217 0.650165,\\\n72  0.540570 0.034950 0.648640,\\\n73  0.546157 0.038954 0.647010,\\\n74  0.551715 0.043136 0.645277,\\\n75  0.557243 0.047331 0.643443,\\\n76  0.562738 0.051545 0.641509,\\\n77  0.568201 0.055778 0.639477,\\\n78  0.573632 0.060028 0.637349,\\\n79  0.579029 0.064296 0.635126,\\\n80  0.584391 0.068579 0.632812,\\\n81  0.589719 0.072878 0.630408,\\\n82  0.595011 0.077190 0.627917,\\\n83  0.600266 0.081516 0.625342,\\\n84  0.605485 0.085854 0.622686,\\\n85  0.610667 0.090204 0.619951,\\\n86  0.615812 0.094564 0.617140,\\\n87  0.620919 0.098934 0.614257,\\\n88  0.625987 0.103312 0.611305,\\\n89  0.631017 0.107699 0.608287,\\\n90  0.636008 0.112092 0.605205,\\\n91  0.640959 0.116492 0.602065,\\\n92  0.645872 0.120898 0.598867,\\\n93  0.650746 0.125309 0.595617,\\\n94  0.655580 0.129725 0.592317,\\\n95  0.660374 0.134144 0.588971,\\\n96  0.665129 0.138566 0.585582,\\\n97  0.669845 0.142992 0.582154,\\\n98  0.674522 0.147419 0.578688,\\\n99  0.679160 0.151848 0.575189,\\\n100 0.683758 0.156278 0.571660,\\\n101 0.688318 0.160709 0.568103,\\\n102 0.692840 0.165141 0.564522,\\\n103 0.697324 0.169573 0.560919,\\\n104 0.701769 0.174005 0.557296,\\\n105 0.706178 0.178437 0.553657,\\\n106 0.710549 0.182868 0.550004,\\\n107 0.714883 0.187299 0.546338,\\\n108 0.719181 0.191729 0.542663,\\\n109 0.723444 0.196158 0.538981,\\\n110 0.727670 0.200586 0.535293,\\\n111 0.731862 0.205013 0.531601,\\\n112 0.736019 0.209439 0.527908,\\\n113 0.740143 0.213864 0.524216,\\\n114 0.744232 0.218288 0.520524,\\\n115 0.748289 0.222711 0.516834,\\\n116 0.752312 0.227133 0.513149,\\\n117 0.756304 0.231555 0.509468,\\\n118 0.760264 0.235976 0.505794,\\\n119 0.764193 0.240396 0.502126,\\\n120 0.768090 0.244817 0.498465,\\\n121 0.771958 0.249237 0.494813,\\\n122 0.775796 0.253658 0.491171,\\\n123 0.779604 0.258078 0.487539,\\\n124 0.783383 0.262500 0.483918,\\\n125 0.787133 0.266922 0.480307,\\\n126 0.790855 0.271345 0.476706,\\\n127 0.794549 0.275770 0.473117,\\\n128 0.798216 0.280197 0.469538,\\\n129 0.801855 0.284626 0.465971,\\\n130 0.805467 0.289057 0.462415,\\\n131 0.809052 0.293491 0.458870,\\\n132 0.812612 0.297928 0.455338,\\\n133 0.816144 0.302368 0.451816,\\\n134 0.819651 0.306812 0.448306,\\\n135 0.823132 0.311261 0.444806,\\\n136 0.826588 0.315714 0.441316,\\\n137 0.830018 0.320172 0.437836,\\\n138 0.833422 0.324635 0.434366,\\\n139 0.836801 0.329105 0.430905,\\\n140 0.840155 0.333580 0.427455,\\\n141 0.843484 0.338062 0.424013,\\\n142 0.846788 0.342551 0.420579,\\\n143 0.850066 0.347048 0.417153,\\\n144 0.853319 0.351553 0.413734,\\\n145 0.856547 0.356066 0.410322,\\\n146 0.859750 0.360588 0.406917,\\\n147 0.862927 0.365119 0.403519,\\\n148 0.866078 0.369660 0.400126,\\\n149 0.869203 0.374212 0.396738,\\\n150 0.872303 0.378774 0.393355,\\\n151 0.875376 0.383347 0.389976,\\\n152 0.878423 0.387932 0.386600,\\\n153 0.881443 0.392529 0.383229,\\\n154 0.884436 0.397139 0.379860,\\\n155 0.887402 0.401762 0.376494,\\\n156 0.890340 0.406398 0.373130,\\\n157 0.893250 0.411048 0.369768,\\\n158 0.896131 0.415712 0.366407,\\\n159 0.898984 0.420392 0.363047,\\\n160 0.901807 0.425087 0.359688,\\\n161 0.904601 0.429797 0.356329,\\\n162 0.907365 0.434524 0.352970,\\\n163 0.910098 0.439268 0.349610,\\\n164 0.912800 0.444029 0.346251,\\\n165 0.915471 0.448807 0.342890,\\\n166 0.918109 0.453603 0.339529,\\\n167 0.920714 0.458417 0.336166,\\\n168 0.923287 0.463251 0.332801,\\\n169 0.925825 0.468103 0.329435,\\\n170 0.928329 0.472975 0.326067,\\\n171 0.930798 0.477867 0.322697,\\\n172 0.933232 0.482780 0.319325,\\\n173 0.935630 0.487712 0.315952,\\\n174 0.937990 0.492667 0.312575,\\\n175 0.940313 0.497642 0.309197,\\\n176 0.942598 0.502639 0.305816,\\\n177 0.944844 0.507658 0.302433,\\\n178 0.947051 0.512699 0.299049,\\\n179 0.949217 0.517763 0.295662,\\\n180 0.951344 0.522850 0.292275,\\\n181 0.953428 0.527960 0.288883,\\\n182 0.955470 0.533093 0.285490,\\\n183 0.957469 0.538250 0.282096,\\\n184 0.959424 0.543431 0.278701,\\\n185 0.961336 0.548636 0.275305,\\\n186 0.963203 0.553865 0.271909,\\\n187 0.965024 0.559118 0.268513,\\\n188 0.966798 0.564396 0.265118,\\\n189 0.968526 0.569700 0.261721,\\\n190 0.970205 0.575028 0.258325,\\\n191 0.971835 0.580382 0.254931,\\\n192 0.973416 0.585761 0.251540,\\\n193 0.974947 0.591165 0.248151,\\\n194 0.976428 0.596595 0.244767,\\\n195 0.977856 0.602051 0.241387,\\\n196 0.979233 0.607532 0.238013,\\\n197 0.980556 0.613039 0.234646,\\\n198 0.981826 0.618572 0.231287,\\\n199 0.983041 0.624131 0.227937,\\\n200 0.984199 0.629718 0.224595,\\\n201 0.985301 0.635330 0.221265,\\\n202 0.986345 0.640969 0.217948,\\\n203 0.987332 0.646633 0.214648,\\\n204 0.988260 0.652325 0.211364,\\\n205 0.989128 0.658043 0.208100,\\\n206 0.989935 0.663787 0.204859,\\\n207 0.990681 0.669558 0.201642,\\\n208 0.991365 0.675355 0.198453,\\\n209 0.991985 0.681179 0.195295,\\\n210 0.992541 0.687030 0.192170,\\\n211 0.993032 0.692907 0.189084,\\\n212 0.993456 0.698810 0.186041,\\\n213 0.993814 0.704741 0.183043,\\\n214 0.994103 0.710698 0.180097,\\\n215 0.994324 0.716681 0.177208,\\\n216 0.994474 0.722691 0.174381,\\\n217 0.994553 0.728728 0.171622,\\\n218 0.994561 0.734791 0.168938,\\\n219 0.994495 0.740880 0.166335,\\\n220 0.994355 0.746995 0.163821,\\\n221 0.994141 0.753137 0.161404,\\\n222 0.993851 0.759304 0.159092,\\\n223 0.993482 0.765499 0.156891,\\\n224 0.993033 0.771720 0.154808,\\\n225 0.992505 0.777967 0.152855,\\\n226 0.991897 0.784239 0.151042,\\\n227 0.991209 0.790537 0.149377,\\\n228 0.990439 0.796859 0.147870,\\\n229 0.989587 0.803205 0.146529,\\\n230 0.988648 0.809579 0.145357,\\\n231 0.987621 0.815978 0.144363,\\\n232 0.986509 0.822401 0.143557,\\\n233 0.985314 0.828846 0.142945,\\\n234 0.984031 0.835315 0.142528,\\\n235 0.982653 0.841812 0.142303,\\\n236 0.981190 0.848329 0.142279,\\\n237 0.979644 0.854866 0.142453,\\\n238 0.977995 0.861432 0.142808,\\\n239 0.976265 0.868016 0.143351,\\\n240 0.974443 0.874622 0.144061,\\\n241 0.972530 0.881250 0.144923,\\\n242 0.970533 0.887896 0.145919,\\\n243 0.968443 0.894564 0.147014,\\\n244 0.966271 0.901249 0.148180,\\\n245 0.964021 0.907950 0.149370,\\\n246 0.961681 0.914672 0.150520,\\\n247 0.959276 0.921407 0.151566,\\\n248 0.956808 0.928152 0.152409,\\\n249 0.954287 0.934908 0.152921,\\\n250 0.951726 0.941671 0.152925,\\\n251 0.949151 0.948435 0.152178,\\\n252 0.946602 0.955190 0.150328,\\\n253 0.944152 0.961916 0.146861,\\\n254 0.941896 0.968590 0.140956,\\\n255 0.940015 0.975158 0.131326)\n" },
    { "prgn", "# line styles for ColorBrewer PRGn (aka PuGn)\n# for use with divering data\n# provides 8 colors with purple low, white middle, and green high\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#762A83' # dark purple\nset style line 2 lt 1 lc rgb '#9970AB' # medium purple\nset style line 3 lt 1 lc rgb '#C2A5CF' # \nset style line 4 lt 1 lc rgb '#E7D4E8' # pale purple\nset style line 5 lt 1 lc rgb '#D9F0D3' # pale green\nset style line 6 lt 1 lc rgb '#A6DBA0' # \nset style line 7 lt 1 lc rgb '#5AAE61' # medium green\nset style line 8 lt 1 lc rgb '#1B7837' # dark green\n\n# palette\nset palette defined ( 0 '#762A83',\\\n    \t    \t      1 '#9970AB',\\\n\t\t      2 '#C2A5CF',\\\n\t\t      3 '#E7D4E8',\\\n\t\t      4 '#D9F0D3',\\\n\t\t      5 '#A6DBA0',\\\n\t\t      6 '#5AAE61',\\\n\t\t      7 '#1B7837' )\n" },
    { "pubu", "# line styles for ColorBrewer PuBu\n# for use with sequential data\n# provides 8 purple-blue colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#FFF7FB' # very light purple-blue\nset style line 2 lt 1 lc rgb '#ECE7F2' # \nset style line 3 lt 1 lc rgb '#D0D1E6' # \nset style line 4 lt 1 lc rgb '#A6BDDB' # light purple-blue\nset style line 5 lt 1 lc rgb '#74A9CF' # \nset style line 6 lt 1 lc rgb '#3690C0' # medium purple-blue\nset style line 7 lt 1 lc rgb '#0570B0' #\nset style line 8 lt 1 lc rgb '#034E7B' # dark purple-blue\n\n# palette\nset palette defined ( 0 '#FFF7FB',\\\n    \t    \t      1 '#ECE7F2',\\\n\t\t      2 '#D0D1E6',\\\n\t\t      3 '#A6BDDB',\\\n\t\t      4 '#74A9CF',\\\n\t\t      5 '#3690C0',\\\n\t\t      6 '#0570B0',\\\n\t\t      7 '#034E7B' )\n" },
    { "pubugn", "# line styles for ColorBrewer PuBuGn\n# for use with sequential data\n# provides 8 purple-blue-green colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#FFF7FB' # very light purple-blue-green\nset style line 2 lt 1 lc rgb '#ECE2F0' # \nset style line 3 lt 1 lc rgb '#D0D1E6' # \nset style line 4 lt 1 lc rgb '#A6BDDB' # light purple-blue-green\nset style line 5 lt 1 lc rgb '#67A9CF' # \nset style line 6 lt 1 lc rgb '#3690C0' # medium purple-blue-green\nset style line 7 lt 1 lc rgb '#02818A' #\nset style line 8 lt 1 lc rgb '#016540' # dark purple-blue-green\n\n# palette\nset palette defined ( 0 '#FFF7FB',\\\n    \t    \t      1 '#ECE7F0',\\\n\t\t      2 '#D0D1E6',\\\n\t\t      3 '#A6BDDB',\\\n\t\t      4 '#67A9CF',\\\n\t\t      5 '#3690C0',\\\n\t\t      6 '#02818A',\\\n\t\t      7 '#016540' )\n" },
    { "puor", "# line styles for ColorBrewer PuOr\n# for use with divering data\n# provides 8 colors with orange low, white middle, and purple high\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#B35806' # dark orange\nset style line 2 lt 1 lc rgb '#E08214' # medium orange\nset style line 3 lt 1 lc rgb '#FDB863' # \nset style line 4 lt 1 lc rgb '#FEE0B6' # pale orange\nset style line 5 lt 1 lc rgb '#D8DAEB' # pale purple\nset style line 6 lt 1 lc rgb '#B2ABD2' # \nset style line 7 lt 1 lc rgb '#8073AC' # medium purple\nset style line 8 lt 1 lc rgb '#542788' # dark purple\n\n# palette\nset palette defined ( 0 '#B35806',\\\n    \t    \t      1 '#E08214',\\\n\t\t      2 '#FDB863',\\\n\t\t      3 '#FEE0B6',\\\n\t\t      4 '#D8DAEB',\\\n\t\t      5 '#B2ABD2',\\\n\t\t      6 '#8073AC',\\\n\t\t      7 '#542788' )\n" },
    { "purd", "# line styles for ColorBrewer PuRd\n# for use with sequential data\n# provides 8 purple-red colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#F7F4F9' # very light purple-red\nset style line 2 lt 1 lc rgb '#E7E1EF' # \nset style line 3 lt 1 lc rgb '#D4B9DA' # \nset style line 4 lt 1 lc rgb '#C994C7' # light purple-red\nset style line 5 lt 1 lc rgb '#DF65B0' # \nset style line 6 lt 1 lc rgb '#E7298A' # medium purple-red\nset style line 7 lt 1 lc rgb '#CE1256' #\nset style line 8 lt 1 lc rgb '#91003F' # dark purple-red\n\n# palette\nset palette defined ( 0 '#F7F4F9',\\\n    \t    \t      1 '#E7E1EF',\\\n\t\t      2 '#D4B9DA',\\\n\t\t      3 '#C994C7',\\\n\t\t      4 '#DF65B0',\\\n\t\t      5 '#E7298A',\\\n\t\t      6 '#CE1256',\\\n\t\t      7 '#91003F' )\n" },
    { "purples", "# line styles for ColorBrewer Purples\n# for use with sequential data\n# provides 8 purple colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#FCFBFD' # very light purple\nset style line 2 lt 1 lc rgb '#EFEDF5' # \nset style line 3 lt 1 lc rgb '#DADAEB' # \nset style line 4 lt 1 lc rgb '#BCBDDC' # light purple\nset style line 5 lt 1 lc rgb '#9E9AC8' # \nset style line 6 lt 1 lc rgb '#807DBA' # medium purple\nset style line 7 lt 1 lc rgb '#6A51A3' #\nset style line 8 lt 1 lc rgb '#4A1486' # dark purple\n\n# palette\nset palette defined ( 0 '#FCFBFD',\\\n    \t    \t      1 '#EFEDF5',\\\n\t\t      2 '#DADAEB',\\\n\t\t      3 '#BCBDDC',\\\n\t\t      4 '#9E9AC8',\\\n\t\t      5 '#807DBA',\\\n\t\t      6 '#6A51A3',\\\n\t\t      7 '#4A1486' )\n" },
    { "rdbu", "# line styles for ColorBrewer RdBu\n# for use with divering data\n# provides 8 colors with red low, white middle, and blue high\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#B2182B' # red\nset style line 2 lt 1 lc rgb '#D6604D' # red-orange\nset style line 3 lt 1 lc rgb '#F4A582' # \nset style line 4 lt 1 lc rgb '#FDDBC7' # pale orange\nset style line 5 lt 1 lc rgb '#D1E5F0' # pale blue\nset style line 6 lt 1 lc rgb '#92C5DE' # \nset style line 7 lt 1 lc rgb '#4393C3' # medium blue\nset style line 8 lt 1 lc rgb '#2166AC' # dark blue\n\n# palette\nset palette defined ( 0 '#B2182B',\\\n    \t    \t      1 '#D6604D',\\\n\t\t      2 '#F4A582',\\\n\t\t      3 '#FDDBC7',\\\n\t\t      4 '#D1E5F0',\\\n\t\t      5 '#92C5DE',\\\n\t\t      6 '#4393C3',\\\n\t\t      7 '#2166AC' )\n" },
    { "rdgy", "# line styles for ColorBrewer RdGy\n# for use with divering data\n# provides 8 colors with red low, white middle, and grey high\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#B2182B' # red\nset style line 2 lt 1 lc rgb '#D6604D' # red-orange\nset style line 3 lt 1 lc rgb '#F4A582' # \nset style line 4 lt 1 lc rgb '#FDDBC7' # pale orange\nset style line 5 lt 1 lc rgb '#E0E0E0' # pale grey\nset style line 6 lt 1 lc rgb '#BABABA' # \nset style line 7 lt 1 lc rgb '#878787' # medium grey\nset style line 8 lt 1 lc rgb '#4D4D4D' # dark grey\n\n# palette\nset palette defined ( 0 '#B2182B',\\\n    \t    \t      1 '#D6604D',\\\n\t\t      2 '#F4A582',\\\n\t\t      3 '#FDDBC7',\\\n\t\t      4 '#E0E0E0',\\\n\t\t      5 '#BABABA',\\\n\t\t      6 '#878787',\\\n\t\t      7 '#4D4D4D' )\n" },
    { "rdpu", "# line styles for ColorBrewer PuRd\n# for use with sequential data\n# provides 8 red-purple colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#FFF7F3' # very light red-purple\nset style line 2 lt 1 lc rgb '#FDE0DD' # \nset style line 3 lt 1 lc rgb '#FCC5C0' # \nset style line 4 lt 1 lc rgb '#FA9FB5' # light red-purple\nset style line 5 lt 1 lc rgb '#F768A1' # \nset style line 6 lt 1 lc rgb '#DD3497' # medium red-purple\nset style line 7 lt 1 lc rgb '#AE017E' #\nset style line 8 lt 1 lc rgb '#7A0177' # dark red-purple\n\n# palette\nset palette defined ( 0 '#FFF7F3',\\\n    \t    \t      1 '#FDE0DD',\\\n\t\t      2 '#FCC5C0',\\\n\t\t      3 '#FA9FB5',\\\n\t\t      4 '#F768A1',\\\n\t\t      5 '#DD3497',\\\n\t\t      6 '#AE017E',\\\n\t\t      7 '#7A0177' )\n" },
    { "rdylbu", "# line styles for ColorBrewer RdYlBu\n# for use with divering data\n# provides 8 colors with red low, pale yellow middle, and blue high\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#D73027' # red\nset style line 2 lt 1 lc rgb '#F46D43' # orange\nset style line 3 lt 1 lc rgb '#FDAE61' # \nset style line 4 lt 1 lc rgb '#FEE090' # pale orange\nset style line 5 lt 1 lc rgb '#E0F3F8' # pale blue\nset style line 6 lt 1 lc rgb '#ABD9E9' # \nset style line 7 lt 1 lc rgb '#74ADD1' # medium blue\nset style line 8 lt 1 lc rgb '#4575B4' # blue\n\n# palette\nset palette defined ( 0 '#D73027',\\\n    \t    \t      1 '#F46D43',\\\n\t\t      2 '#FDAE61',\\\n\t\t      3 '#FEE090',\\\n\t\t      4 '#E0F3F8',\\\n\t\t      5 '#ABD9E9',\\\n\t\t      6 '#74ADD1',\\\n\t\t      7 '#4575B4' )\n" },
    { "rdylgn", "# line styles for ColorBrewer RdYlGn\n# for use with divering data\n# provides 8 colors with red low, pale yellow middle, and green high\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#D73027' # red\nset style line 2 lt 1 lc rgb '#F46D43' # orange\nset style line 3 lt 1 lc rgb '#FDAE61' # pale orange\nset style line 4 lt 1 lc rgb '#FEE08B' # pale yellow-orange\nset style line 5 lt 1 lc rgb '#D9EF8B' # pale yellow-green\nset style line 6 lt 1 lc rgb '#A6D96A' # pale green\nset style line 7 lt 1 lc rgb '#66BD63' # medium green\nset style line 8 lt 1 lc rgb '#1A9850' # green\n\n# palette\nset palette defined ( 0 '#D73027',\\\n    \t    \t      1 '#F46D43',\\\n\t\t      2 '#FDAE61',\\\n\t\t      3 '#FEE08B',\\\n\t\t      4 '#D9EF8B',\\\n\t\t      5 '#A6D96A',\\\n\t\t      6 '#66BD63',\\\n\t\t      7 '#1A9850' )\n" },
    { "reds", "# line styles for ColorBrewer Reds\n# for use with sequential data\n# provides 8 red colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#FFF5F0' # very light red\nset style line 2 lt 1 lc rgb '#FEE0D2' # \nset style line 3 lt 1 lc rgb '#FCBBA1' # \nset style line 4 lt 1 lc rgb '#FC9272' # light red\nset style line 5 lt 1 lc rgb '#FB6A4A' # \nset style line 6 lt 1 lc rgb '#EF3B2C' # medium red\nset style line 7 lt 1 lc rgb '#CB181D' #\nset style line 8 lt 1 lc rgb '#99000D' # dark red\n\n# palette\nset palette defined ( 0 '#FFF5F0',\\\n    \t    \t      1 '#FEE0D2',\\\n\t\t      2 '#FCBBA1',\\\n\t\t      3 '#FC9272',\\\n\t\t      4 '#FB6A4A',\\\n\t\t      5 '#EF3B2C',\\\n\t\t      6 '#CB181D',\\\n\t\t      7 '#99000D' )\n" },
    { "sand", "# Sequential palette in sand colors\n# From G. N. Peterson, Cartographer's toolkit, PertersonGIS, 2012\n\n# line styles\nset style line  1 lt 1 lc rgb '#604860' # dark brown\nset style line  2 lt 1 lc rgb '#784860' # brown\nset style line  3 lt 1 lc rgb '#a86060' # brown\nset style line  4 lt 1 lc rgb '#c07860' # light brown\nset style line  5 lt 1 lc rgb '#f0a848' # orange\nset style line  6 lt 1 lc rgb '#f8ca8c' # orange\nset style line  7 lt 1 lc rgb '#feecae' # yellow\nset style line  8 lt 1 lc rgb '#fff4c2' # yellow\nset style line  9 lt 1 lc rgb '#fff7db' # light yellow\nset style line 10 lt 1 lc rgb '#fffcf6' # white\n\n# palette\nset palette defined (\\\n0 '#604860',\\\n1 '#784860',\\\n2 '#a86060',\\\n3 '#c07860',\\\n4 '#f0a848',\\\n5 '#f8ca8c',\\\n6 '#feecae',\\\n7 '#fff4c2',\\\n8 '#fff7db',\\\n9 '#fffcf6')\n" },
    { "set1", "# line styles for ColorBrewer Set1\n# for use with qualitative/categorical data\n# provides 8 easy-to-name colors\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#E41A1C' # red\nset style line 2 lt 1 lc rgb '#377EB8' # blue\nset style line 3 lt 1 lc rgb '#4DAF4A' # green\nset style line 4 lt 1 lc rgb '#984EA3' # purple\nset style line 5 lt 1 lc rgb '#FF7F00' # orange\nset style line 6 lt 1 lc rgb '#FFFF33' # yellow\nset style line 7 lt 1 lc rgb '#A65628' # brown\nset style line 8 lt 1 lc rgb '#F781BF' # pink\n\n# palette\nset palette maxcolors 8\nset palette defined ( 0 '#E41A1C',\\\n    \t    \t      1 '#377EB8',\\\n\t\t      2 '#4DAF4A',\\\n\t\t      3 '#984EA3',\\\n\t\t      4 '#FF7F00',\\\n\t\t      5 '#FFFF33',\\\n\t\t      6 '#A65628',\\\n\t\t      7 '#F781BF' )\n" },
    { "set2", "# line styles for ColorBrewer Set3\n# for use with qualitative/categorical data\n# provides 8 hard-to-name colors\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#66C2A5' # teal\nset style line 2 lt 1 lc rgb '#FC8D62' # orange\nset style line 3 lt 1 lc rgb '#8DA0CB' # lilac\nset style line 4 lt 1 lc rgb '#E78AC3' # magentat\nset style line 5 lt 1 lc rgb '#A6D854' # lime green\nset style line 6 lt 1 lc rgb '#FFD92F' # banana\nset style line 7 lt 1 lc rgb '#E5C494' # tan\nset style line 8 lt 1 lc rgb '#B3B3B3' # grey\n\n# palette\nset palette maxcolors 8\nset palette defined ( 0 '#66C2A5',\\\n    \t    \t      1 '#FC8D62',\\\n\t\t      2 '#8DA0CB',\\\n\t\t      3 '#E78AC3',\\\n\t\t      4 '#A6D854',\\\n\t\t      5 '#FFD92F',\\\n\t\t      6 '#E5C494',\\\n\t\t      7 '#B3B3B3' )\n" },
    { "set3", "# line styles for ColorBrewer Set3\n# for use with qualitative/categorical data\n# provides 8 colors that are more saturated than the Pastels but less so than Set2\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#8DD3C7' # teal\nset style line 2 lt 1 lc rgb '#FFFFB3' # banana\nset style line 3 lt 1 lc rgb '#BEBADA' # lilac\nset style line 4 lt 1 lc rgb '#FB8072' # red\nset style line 5 lt 1 lc rgb '#80B1D3' # steel blue\nset style line 6 lt 1 lc rgb '#FDB462' # adobe orange\nset style line 7 lt 1 lc rgb '#B3DE69' # lime green\nset style line 8 lt 1 lc rgb '#FCCDE5' # mauve\n\n# palette\nset palette maxcolors 8\nset palette defined ( 0 '#8DD3C7',\\\n    \t    \t      1 '#FFFFB3',\\\n\t\t      2 '#BEBADA',\\\n\t\t      3 '#FB8072',\\\n\t\t      4 '#80B1D3',\\\n\t\t      5 '#FDB462',\\\n\t\t      6 '#B3DE69',\\\n\t\t      7 '#FCCDE5' )\n" },
    { "spectral", "# line styles for ColorBrewer Spectral\n# for use with divering data\n# provides 8 rainbow colors with red low, pale yellow middle, and blue high\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#D53E4F' # red\nset style line 2 lt 1 lc rgb '#F46D43' # orange\nset style line 3 lt 1 lc rgb '#FDAE61' # pale orange\nset style line 4 lt 1 lc rgb '#FEE08B' # pale yellow-orange\nset style line 5 lt 1 lc rgb '#E6F598' # pale yellow-green\nset style line 6 lt 1 lc rgb '#ABDDA4' # pale green\nset style line 7 lt 1 lc rgb '#66C2A5' # green\nset style line 8 lt 1 lc rgb '#3288BD' # blue\n\n# palette\nset palette defined ( 0 '#D53E4F',\\\n    \t    \t      1 '#F46D43',\\\n\t\t      2 '#FDAE61',\\\n\t\t      3 '#FEE08B',\\\n\t\t      4 '#E6F598',\\\n\t\t      5 '#ABDDA4',\\\n\t\t      6 '#66C2A5',\\\n\t\t      7 '#3288BD' )\n" },
    { "viridis", "# New matplotlib colormaps by Nathaniel J. Smith, Stefan van der Walt,\n# and (in the case of viridis) Eric Firing.\n#\n# This file and the colormaps in it are released under the CC0 license /\n# public domain dedication. We would appreciate credit if you use or\n# redistribute these colormaps, but do not impose any legal restrictions.\n#\n# To the extent possible under law, the persons who associated CC0 with\n# mpl-colormaps have waived all copyright and related or neighboring rights\n# to mpl-colormaps.\n#\n# You should have received a copy of the CC0 legalcode along with this\n# work.  If not, see <http://creativecommons.org/publicdomain/zero/1.0/>.\n\n#https://github.com/BIDS/colormap/blob/master/colormaps.py\n\n\n# line styles\nset style line  1 lt 1 lc rgb '#440154' # dark purple\nset style line  2 lt 1 lc rgb '#472c7a' # purple\nset style line  3 lt 1 lc rgb '#3b518b' # blue\nset style line  4 lt 1 lc rgb '#2c718e' # blue\nset style line  5 lt 1 lc rgb '#21908d' # blue-green\nset style line  6 lt 1 lc rgb '#27ad81' # green\nset style line  7 lt 1 lc rgb '#5cc863' # green\nset style line  8 lt 1 lc rgb '#aadc32' # lime green\nset style line  9 lt 1 lc rgb '#fde725' # yellow\n\n\n# palette\nset palette defined (\\\n0   0.267004 0.004874 0.329415,\\\n1   0.268510 0.009605 0.335427,\\\n2   0.269944 0.014625 0.341379,\\\n3   0.271305 0.019942 0.347269,\\\n4   0.272594 0.025563 0.353093,\\\n5   0.273809 0.031497 0.358853,\\\n6   0.274952 0.037752 0.364543,\\\n7   0.276022 0.044167 0.370164,\\\n8   0.277018 0.050344 0.375715,\\\n9   0.277941 0.056324 0.381191,\\\n10  0.278791 0.062145 0.386592,\\\n11  0.279566 0.067836 0.391917,\\\n12  0.280267 0.073417 0.397163,\\\n13  0.280894 0.078907 0.402329,\\\n14  0.281446 0.084320 0.407414,\\\n15  0.281924 0.089666 0.412415,\\\n16  0.282327 0.094955 0.417331,\\\n17  0.282656 0.100196 0.422160,\\\n18  0.282910 0.105393 0.426902,\\\n19  0.283091 0.110553 0.431554,\\\n20  0.283197 0.115680 0.436115,\\\n21  0.283229 0.120777 0.440584,\\\n22  0.283187 0.125848 0.444960,\\\n23  0.283072 0.130895 0.449241,\\\n24  0.282884 0.135920 0.453427,\\\n25  0.282623 0.140926 0.457517,\\\n26  0.282290 0.145912 0.461510,\\\n27  0.281887 0.150881 0.465405,\\\n28  0.281412 0.155834 0.469201,\\\n29  0.280868 0.160771 0.472899,\\\n30  0.280255 0.165693 0.476498,\\\n31  0.279574 0.170599 0.479997,\\\n32  0.278826 0.175490 0.483397,\\\n33  0.278012 0.180367 0.486697,\\\n34  0.277134 0.185228 0.489898,\\\n35  0.276194 0.190074 0.493001,\\\n36  0.275191 0.194905 0.496005,\\\n37  0.274128 0.199721 0.498911,\\\n38  0.273006 0.204520 0.501721,\\\n39  0.271828 0.209303 0.504434,\\\n40  0.270595 0.214069 0.507052,\\\n41  0.269308 0.218818 0.509577,\\\n42  0.267968 0.223549 0.512008,\\\n43  0.266580 0.228262 0.514349,\\\n44  0.265145 0.232956 0.516599,\\\n45  0.263663 0.237631 0.518762,\\\n46  0.262138 0.242286 0.520837,\\\n47  0.260571 0.246922 0.522828,\\\n48  0.258965 0.251537 0.524736,\\\n49  0.257322 0.256130 0.526563,\\\n50  0.255645 0.260703 0.528312,\\\n51  0.253935 0.265254 0.529983,\\\n52  0.252194 0.269783 0.531579,\\\n53  0.250425 0.274290 0.533103,\\\n54  0.248629 0.278775 0.534556,\\\n55  0.246811 0.283237 0.535941,\\\n56  0.244972 0.287675 0.537260,\\\n57  0.243113 0.292092 0.538516,\\\n58  0.241237 0.296485 0.539709,\\\n59  0.239346 0.300855 0.540844,\\\n60  0.237441 0.305202 0.541921,\\\n61  0.235526 0.309527 0.542944,\\\n62  0.233603 0.313828 0.543914,\\\n63  0.231674 0.318106 0.544834,\\\n64  0.229739 0.322361 0.545706,\\\n65  0.227802 0.326594 0.546532,\\\n66  0.225863 0.330805 0.547314,\\\n67  0.223925 0.334994 0.548053,\\\n68  0.221989 0.339161 0.548752,\\\n69  0.220057 0.343307 0.549413,\\\n70  0.218130 0.347432 0.550038,\\\n71  0.216210 0.351535 0.550627,\\\n72  0.214298 0.355619 0.551184,\\\n73  0.212395 0.359683 0.551710,\\\n74  0.210503 0.363727 0.552206,\\\n75  0.208623 0.367752 0.552675,\\\n76  0.206756 0.371758 0.553117,\\\n77  0.204903 0.375746 0.553533,\\\n78  0.203063 0.379716 0.553925,\\\n79  0.201239 0.383670 0.554294,\\\n80  0.199430 0.387607 0.554642,\\\n81  0.197636 0.391528 0.554969,\\\n82  0.195860 0.395433 0.555276,\\\n83  0.194100 0.399323 0.555565,\\\n84  0.192357 0.403199 0.555836,\\\n85  0.190631 0.407061 0.556089,\\\n86  0.188923 0.410910 0.556326,\\\n87  0.187231 0.414746 0.556547,\\\n88  0.185556 0.418570 0.556753,\\\n89  0.183898 0.422383 0.556944,\\\n90  0.182256 0.426184 0.557120,\\\n91  0.180629 0.429975 0.557282,\\\n92  0.179019 0.433756 0.557430,\\\n93  0.177423 0.437527 0.557565,\\\n94  0.175841 0.441290 0.557685,\\\n95  0.174274 0.445044 0.557792,\\\n96  0.172719 0.448791 0.557885,\\\n97  0.171176 0.452530 0.557965,\\\n98  0.169646 0.456262 0.558030,\\\n99  0.168126 0.459988 0.558082,\\\n100 0.166617 0.463708 0.558119,\\\n101 0.165117 0.467423 0.558141,\\\n102 0.163625 0.471133 0.558148,\\\n103 0.162142 0.474838 0.558140,\\\n104 0.160665 0.478540 0.558115,\\\n105 0.159194 0.482237 0.558073,\\\n106 0.157729 0.485932 0.558013,\\\n107 0.156270 0.489624 0.557936,\\\n108 0.154815 0.493313 0.557840,\\\n109 0.153364 0.497000 0.557724,\\\n110 0.151918 0.500685 0.557587,\\\n111 0.150476 0.504369 0.557430,\\\n112 0.149039 0.508051 0.557250,\\\n113 0.147607 0.511733 0.557049,\\\n114 0.146180 0.515413 0.556823,\\\n115 0.144759 0.519093 0.556572,\\\n116 0.143343 0.522773 0.556295,\\\n117 0.141935 0.526453 0.555991,\\\n118 0.140536 0.530132 0.555659,\\\n119 0.139147 0.533812 0.555298,\\\n120 0.137770 0.537492 0.554906,\\\n121 0.136408 0.541173 0.554483,\\\n122 0.135066 0.544853 0.554029,\\\n123 0.133743 0.548535 0.553541,\\\n124 0.132444 0.552216 0.553018,\\\n125 0.131172 0.555899 0.552459,\\\n126 0.129933 0.559582 0.551864,\\\n127 0.128729 0.563265 0.551229,\\\n128 0.127568 0.566949 0.550556,\\\n129 0.126453 0.570633 0.549841,\\\n130 0.125394 0.574318 0.549086,\\\n131 0.124395 0.578002 0.548287,\\\n132 0.123463 0.581687 0.547445,\\\n133 0.122606 0.585371 0.546557,\\\n134 0.121831 0.589055 0.545623,\\\n135 0.121148 0.592739 0.544641,\\\n136 0.120565 0.596422 0.543611,\\\n137 0.120092 0.600104 0.542530,\\\n138 0.119738 0.603785 0.541400,\\\n139 0.119512 0.607464 0.540218,\\\n140 0.119423 0.611141 0.538982,\\\n141 0.119483 0.614817 0.537692,\\\n142 0.119699 0.618490 0.536347,\\\n143 0.120081 0.622161 0.534946,\\\n144 0.120638 0.625828 0.533488,\\\n145 0.121380 0.629492 0.531973,\\\n146 0.122312 0.633153 0.530398,\\\n147 0.123444 0.636809 0.528763,\\\n148 0.124780 0.640461 0.527068,\\\n149 0.126326 0.644107 0.525311,\\\n150 0.128087 0.647749 0.523491,\\\n151 0.130067 0.651384 0.521608,\\\n152 0.132268 0.655014 0.519661,\\\n153 0.134692 0.658636 0.517649,\\\n154 0.137339 0.662252 0.515571,\\\n155 0.140210 0.665859 0.513427,\\\n156 0.143303 0.669459 0.511215,\\\n157 0.146616 0.673050 0.508936,\\\n158 0.150148 0.676631 0.506589,\\\n159 0.153894 0.680203 0.504172,\\\n160 0.157851 0.683765 0.501686,\\\n161 0.162016 0.687316 0.499129,\\\n162 0.166383 0.690856 0.496502,\\\n163 0.170948 0.694384 0.493803,\\\n164 0.175707 0.697900 0.491033,\\\n165 0.180653 0.701402 0.488189,\\\n166 0.185783 0.704891 0.485273,\\\n167 0.191090 0.708366 0.482284,\\\n168 0.196571 0.711827 0.479221,\\\n169 0.202219 0.715272 0.476084,\\\n170 0.208030 0.718701 0.472873,\\\n171 0.214000 0.722114 0.469588,\\\n172 0.220124 0.725509 0.466226,\\\n173 0.226397 0.728888 0.462789,\\\n174 0.232815 0.732247 0.459277,\\\n175 0.239374 0.735588 0.455688,\\\n176 0.246070 0.738910 0.452024,\\\n177 0.252899 0.742211 0.448284,\\\n178 0.259857 0.745492 0.444467,\\\n179 0.266941 0.748751 0.440573,\\\n180 0.274149 0.751988 0.436601,\\\n181 0.281477 0.755203 0.432552,\\\n182 0.288921 0.758394 0.428426,\\\n183 0.296479 0.761561 0.424223,\\\n184 0.304148 0.764704 0.419943,\\\n185 0.311925 0.767822 0.415586,\\\n186 0.319809 0.770914 0.411152,\\\n187 0.327796 0.773980 0.406640,\\\n188 0.335885 0.777018 0.402049,\\\n189 0.344074 0.780029 0.397381,\\\n190 0.352360 0.783011 0.392636,\\\n191 0.360741 0.785964 0.387814,\\\n192 0.369214 0.788888 0.382914,\\\n193 0.377779 0.791781 0.377939,\\\n194 0.386433 0.794644 0.372886,\\\n195 0.395174 0.797475 0.367757,\\\n196 0.404001 0.800275 0.362552,\\\n197 0.412913 0.803041 0.357269,\\\n198 0.421908 0.805774 0.351910,\\\n199 0.430983 0.808473 0.346476,\\\n200 0.440137 0.811138 0.340967,\\\n201 0.449368 0.813768 0.335384,\\\n202 0.458674 0.816363 0.329727,\\\n203 0.468053 0.818921 0.323998,\\\n204 0.477504 0.821444 0.318195,\\\n205 0.487026 0.823929 0.312321,\\\n206 0.496615 0.826376 0.306377,\\\n207 0.506271 0.828786 0.300362,\\\n208 0.515992 0.831158 0.294279,\\\n209 0.525776 0.833491 0.288127,\\\n210 0.535621 0.835785 0.281908,\\\n211 0.545524 0.838039 0.275626,\\\n212 0.555484 0.840254 0.269281,\\\n213 0.565498 0.842430 0.262877,\\\n214 0.575563 0.844566 0.256415,\\\n215 0.585678 0.846661 0.249897,\\\n216 0.595839 0.848717 0.243329,\\\n217 0.606045 0.850733 0.236712,\\\n218 0.616293 0.852709 0.230052,\\\n219 0.626579 0.854645 0.223353,\\\n220 0.636902 0.856542 0.216620,\\\n221 0.647257 0.858400 0.209861,\\\n222 0.657642 0.860219 0.203082,\\\n223 0.668054 0.861999 0.196293,\\\n224 0.678489 0.863742 0.189503,\\\n225 0.688944 0.865448 0.182725,\\\n226 0.699415 0.867117 0.175971,\\\n227 0.709898 0.868751 0.169257,\\\n228 0.720391 0.870350 0.162603,\\\n229 0.730889 0.871916 0.156029,\\\n230 0.741388 0.873449 0.149561,\\\n231 0.751884 0.874951 0.143228,\\\n232 0.762373 0.876424 0.137064,\\\n233 0.772852 0.877868 0.131109,\\\n234 0.783315 0.879285 0.125405,\\\n235 0.793760 0.880678 0.120005,\\\n236 0.804182 0.882046 0.114965,\\\n237 0.814576 0.883393 0.110347,\\\n238 0.824940 0.884720 0.106217,\\\n239 0.835270 0.886029 0.102646,\\\n240 0.845561 0.887322 0.099702,\\\n241 0.855810 0.888601 0.097452,\\\n242 0.866013 0.889868 0.095953,\\\n243 0.876168 0.891125 0.095250,\\\n244 0.886271 0.892374 0.095374,\\\n245 0.896320 0.893616 0.096335,\\\n246 0.906311 0.894855 0.098125,\\\n247 0.916242 0.896091 0.100717,\\\n248 0.926106 0.897330 0.104071,\\\n249 0.935904 0.898570 0.108131,\\\n250 0.945636 0.899815 0.112838,\\\n251 0.955300 0.901065 0.118128,\\\n252 0.964894 0.902323 0.123941,\\\n253 0.974417 0.903590 0.130215,\\\n254 0.983868 0.904867 0.136897,\\\n255 0.993248 0.906157 0.143936)\n" },
    { "whylrd", "# line styles\nset style line 1 lt 1 lc rgb '#ffffff' # white\nset style line 2 lt 1 lc rgb '#ffee00' # \nset style line 3 lt 1 lc rgb '#ff7000' # yellow\nset style line 4 lt 1 lc rgb '#ee0000' #\nset style line 5 lt 1 lc rgb '#7f0000' # red\n\n# palette\nset palette defined ( \\\n    0 '#ffffff', \\\n    1 '#ffee00', \\\n    2 '#ff7000', \\\n    3 '#ee0000', \\\n    4 '#7f0000')\n" },
    { "ylgn", "# line styles for ColorBrewer YlGn\n# for use with sequential data\n# provides 8 yellow-green colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#FFFFE5' # very light yellow-green\nset style line 2 lt 1 lc rgb '#F7FCB9' # \nset style line 3 lt 1 lc rgb '#D9F0A3' # \nset style line 4 lt 1 lc rgb '#ADDD8E' # light yellow-green\nset style line 5 lt 1 lc rgb '#78C679' # \nset style line 6 lt 1 lc rgb '#41AB5D' # medium yellow-green\nset style line 7 lt 1 lc rgb '#238443' #\nset style line 8 lt 1 lc rgb '#005A32' # dark yellow-green\n\n# palette\nset palette defined ( 0 '#FFFFE5',\\\n    \t    \t      1 '#F7FCB9',\\\n\t\t      2 '#D9F0A3',\\\n\t\t      3 '#ADDD8E',\\\n\t\t      4 '#78C679',\\\n\t\t      5 '#41AB5D',\\\n\t\t      6 '#238443',\\\n\t\t      7 '#005A32' )\n" },
    { "ylgnbu", "# line styles for ColorBrewer YlGnBu\n# for use with sequential data\n# provides 8 yellow-green-blue colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#FFFFD9' # very light yellow-green-blue\nset style line 2 lt 1 lc rgb '#EDF8B1' # \nset style line 3 lt 1 lc rgb '#C7E9B4' # \nset style line 4 lt 1 lc rgb '#7FCDBB' # light yellow-green-blue\nset style line 5 lt 1 lc rgb '#41B6C4' # \nset style line 6 lt 1 lc rgb '#1D91C0' # medium yellow-green-blue\nset style line 7 lt 1 lc rgb '#225EA8' #\nset style line 8 lt 1 lc rgb '#0C2C84' # dark yellow-green-blue\n\n# palette\nset palette defined ( 0 '#FFFFD9',\\\n    \t    \t      1 '#EDF8B1',\\\n\t\t      2 '#C7E9B4',\\\n\t\t      3 '#7FCDBB',\\\n\t\t      4 '#41B6C4',\\\n\t\t      5 '#1D91C0',\\\n\t\t      6 '#225EA8',\\\n\t\t      7 '#0C2C84' )\n" },
    { "ylorbr", "# line styles for ColorBrewer YlOrBr\n# for use with sequential data\n# provides 8 yellow-orange-brown colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#FFFFE5' # very light yellow-orange-brown\nset style line 2 lt 1 lc rgb '#FFF7BC' # \nset style line 3 lt 1 lc rgb '#FEE391' # \nset style line 4 lt 1 lc rgb '#FEC44F' # light yellow-orange-brown\nset style line 5 lt 1 lc rgb '#FE9929' # \nset style line 6 lt 1 lc rgb '#EC7014' # medium yellow-orange-brown\nset style line 7 lt 1 lc rgb '#CC4C02' #\nset style line 8 lt 1 lc rgb '#CC4C02' # dark yellow-orange-brown\n\n# palette\nset palette defined ( 0 '#FFFFE5',\\\n    \t    \t      1 '#FFF7BC',\\\n\t\t      2 '#FEE391',\\\n\t\t      3 '#FEC44F',\\\n\t\t      4 '#FE9929',\\\n\t\t      5 '#EC7014',\\\n\t\t      6 '#CC4C02',\\\n\t\t      7 '#8C2D04' )\n" },
    { "ylorrd", "# line styles for ColorBrewer YlOrRd\n# for use with sequential data\n# provides 8 yellow-orange-red colors of increasing saturation\n# compatible with gnuplot >=4.2\n# author: Anna Schneider\n\n# line styles\nset style line 1 lt 1 lc rgb '#FFFFCC' # very light yellow-orange-red\nset style line 2 lt 1 lc rgb '#FFEDA0' # \nset style line 3 lt 1 lc rgb '#FED976' # light yellow-orange-red\nset style line 4 lt 1 lc rgb '#FEB24C' # \nset style line 5 lt 1 lc rgb '#FD8D3C' # \nset style line 6 lt 1 lc rgb '#FC4E2A' # medium yellow-orange-red\nset style line 7 lt 1 lc rgb '#E31A1C' #\nset style line 8 lt 1 lc rgb '#B10026' # dark yellow-orange-red\n\n# palette\nset palette defined ( 0 '#FFFFCC',\\\n    \t    \t      1 '#FFEDA0',\\\n\t\t      2 '#FED976',\\\n\t\t      3 '#FEB24C',\\\n\t\t      4 '#FD8D3C',\\\n\t\t      5 '#FC4E2A',\\\n\t\t      6 '#E31A1C',\\\n\t\t      7 '#B10026' )\n" },
    { "ylrd", "# line styles\nset style line 1 lt 1 lc rgb '#ffee00' # yellow \nset style line 2 lt 1 lc rgb '#ff7000' #\nset style line 3 lt 1 lc rgb '#ee0000' #\nset style line 4 lt 1 lc rgb '#7f0000' # red\n\n# palette\nset palette defined ( \\\n    0 '#ffee00', \\\n    1 '#ff7000', \\\n    2 '#ee0000', \\\n    3 '#7f0000')\n" },
};

} // namespace sciplot

